from datetime import datetime, timezone
from decimal import Decimal, InvalidOperation
from PySide6.QtWidgets import QWidget, QStyledItemDelegate, QLineEdit, QDateTimeEdit, QTreeView, QComboBox
from PySide6.QtCore import Qt, QModelIndex, QEvent, QLocale, QDateTime, QDate, QTime, QTimeZone
from PySide6.QtGui import QDoubleValidator, QBrush, QKeyEvent
from jal.constants import CustomColor, Setup
from jal.widgets.reference_selector import AssetSelector, PeerSelector, CategorySelector, TagSelector
from jal.db.db import JalModel
from jal.db.helpers import localize_decimal, delocalize_decimal
from jal.db.account import JalAccount
from jal.widgets.icons import JalIcon


# ----------------------------------------------------------------------------------------------------------------------
# Base class to provide different delegates for WidgetDataMappers in operations widgets
# Separate delegate class is subclassed for every operation widget with own definition of self.delegates for columns
class WidgetMapperDelegateBase(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.timestamp_delegate = TimestampDelegate()
        self.decimal_long_delegate = FloatDelegate(2, allow_tail=True)
        self.decimal_delegate = FloatDelegate(2)
        self.symbol_delegate = SymbolDelegate()
        self.default = QStyledItemDelegate()

        self.delegates = {}

    def get_delegate(self, index):
        column_name = index.model().record().fieldName(index.column())
        try:
            delegate = self.delegates[column_name]
        except KeyError:
            delegate = self.default
        return delegate

    def paint(self, painter, option, index):
        delegate = self.get_delegate(index)
        delegate.paint(painter, option, index)

    def createEditor(self, aParent, option, index):
        delegate = self.get_delegate(index)
        delegate.createEditor(aParent, option, index)

    def setEditorData(self, editor, index):
        delegate = self.get_delegate(index)
        delegate.setEditorData(editor, index)

    def setModelData(self, editor, model, index):
        delegate = self.get_delegate(index)
        delegate.setModelData(editor, model, index)


# ----------------------------------------------------------------------------------------------------------------------
# Custom DateTimeEdit that is able to reset value to None
# i.e. it handles keypress '-', sets timestamp to 0 that effectively cleans the field
class DateTimeEditWithReset(QDateTimeEdit):
    def __init__(self, parent):
        super().__init__(parent)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_Minus:
            self.setDateTime(QDateTime(QDate(1970, 1, 1), QTime(0, 0, 0), Qt.UTC))  # = 0 timestamp
        super().keyPressEvent(event)



# ----------------------------------------------------------------------------------------------------------------------
# QTreeView doesn't draw grid lines and have no normal method to implement it
# So the purpose of this delegate is solely to draw dotted box around report cell
class GridLinesDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        self._parent = parent
        super().__init__(parent=parent)

    def paint_grid(self, painter, option, index):
        if issubclass(type(self._parent), QTreeView):  # Extra code for tree views - to draw grid lines
            painter.save()
            pen = painter.pen()
            pen.setWidth(1)
            pen.setStyle(Qt.DotLine)
            pen.setColor(Qt.GlobalColor.lightGray)
            painter.setPen(pen)
            painter.drawRect(option.rect)
            painter.restore()

    def paint(self, painter, option, index):
        self.paint_grid(painter, option, index)
        super().paint(painter, option, index)


# ----------------------------------------------------------------------------------------------------------------------
# Delegate to convert timestamp from unix-time to QDateTime and display it according to the given format
class TimestampDelegate(GridLinesDelegate):
    def __init__(self, display_format='%d/%m/%Y %H:%M:%S', parent=None):
        super().__init__(parent=parent)
        self._parent = parent
        self._format = display_format

    def displayText(self, value, locale):
        if isinstance(value, str):  # int value comes here in form of string in case of SQL aggregate function results
            try:
                value = int(value)
            except ValueError:
                return self.tr("<invalid>")
        text = datetime.fromtimestamp(value, tz=timezone.utc).strftime(self._format) if value else ''
        return text

    def createEditor(self, aParent, option, index):
        editor = DateTimeEditWithReset(aParent)
        editor.setTimeSpec(Qt.UTC)
        if 'H' in self._format:  # we have hours and need DataTime editor to edit it
            editor.setDisplayFormat("dd/MM/yyyy hh:mm:ss")
        else:
            editor.setDisplayFormat("dd/MM/yyyy")
        return editor

    def setEditorData(self, editor, index):
        timestamp = index.model().data(index, Qt.EditRole)
        if timestamp == '':
            QStyledItemDelegate.setEditorData(self, editor, index)
        else:
            editor.setDateTime(QDateTime.fromSecsSinceEpoch(timestamp, QTimeZone(0)))

    def setModelData(self, editor, model, index):
        timestamp = editor.dateTime().toSecsSinceEpoch()
        model.setData(index, timestamp)


# -----------------------------------------------------------------------------------------------------------------------
# Delegate for float numbers formatting
# By default has 6 decimal places that may be controlled with 'tolerance' parameter
# 'allow_tail' - display more digits for numbers that have more digits than 'tolerance'
#                and only 'tolerance' digits otherwise
# 'colors' - make Green/Red background for positive/negative values
# 'percent' - multiply values by 100 in order to display percents
# 'empty_zero' - display nothing instead of number 0
class FloatDelegate(GridLinesDelegate):
    DEFAULT_TOLERANCE = 6

    def __init__(self, tolerance=None, allow_tail=True, colors=False, percent=False, empty_zero=False, parent=None):
        self._parent = parent
        super().__init__(parent=parent)
        try:
            self._tolerance = int(tolerance)
        except (ValueError, TypeError):
            self._tolerance = self.DEFAULT_TOLERANCE
        self._allow_tail = allow_tail
        self._colors = colors
        self._color = None
        self._percent = percent
        self._empty_zero = empty_zero
        self._validator = QDoubleValidator()
        self._validator.setLocale(QLocale().system())

    def displayText(self, value, locale):
        try:
            amount = Decimal(value)
        except (TypeError, ValueError, InvalidOperation):
            amount = None
        if amount is None or amount.is_nan():
            return localize_decimal(amount)
        if self._percent:
            amount *= Decimal('100')
        if amount > Decimal('0'):
            self._color = CustomColor.LightGreen
        elif amount < Decimal('0'):
            self._color = CustomColor.LightRed
        else:
            self._color = None
            if self._empty_zero:
                return ''
        decimal_places = -amount.normalize().as_tuple().exponent
        decimal_places = decimal_places if self._allow_tail and (decimal_places > self._tolerance) else self._tolerance
        return localize_decimal(amount, decimal_places)

    # this is required when edit operation is called from QTableView
    def createEditor(self, aParent, option, index):
        float_editor = QLineEdit(aParent)
        float_editor.setValidator(self._validator)
        return float_editor

    def setEditorData(self, editor, index):
        try:
            amount = Decimal(index.model().data(index, Qt.EditRole))
        except (InvalidOperation, TypeError):   # Set to zero if we have None in database
            amount = Decimal('0')
        formatted_text = localize_decimal(amount, precision=self._tolerance, percent=self._percent)
        if self._allow_tail:
            full_text = localize_decimal(amount, percent=self._percent)
            if len(full_text) > len(formatted_text):
                formatted_text = full_text
        editor.setText(formatted_text)

    def setModelData(self, editor, model, index):
        model.setData(index, str(delocalize_decimal(editor.text(), percent=self._percent)))

    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        option.displayAlignment = Qt.AlignRight
        if self._colors and self._color is not None:
            option.backgroundBrush = QBrush(self._color)


# ----------------------------------------------------------------------------------------------------------------------
# Delegate to apply currency filter for AssetSelector widgets based on current account
class SymbolDelegate(QStyledItemDelegate):
    def setEditorData(self, editor, index):
        account_currency = JalAccount(index.model().data(index.sibling(index.row(),
                                                                       index.model().fieldIndex('account_id')),
                                                         Qt.EditRole)).currency()
        editor.setFilterValue(account_currency)
        QStyledItemDelegate.setEditorData(self, editor, index)


# ----------------------------------------------------------------------------------------------------------------------
# Toggle True/False by mouse click and display status by relevant icon
class BoolDelegate(GridLinesDelegate):
    def __init__(self, parent=None):
        self._parent = parent
        super().__init__(parent=parent)

    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        option.displayAlignment = Qt.AlignCenter

    def paint(self, painter, option, index):
        value = index.model().data(index, Qt.DisplayRole)
        painter.save()
        if value:
            icon = JalIcon[JalIcon.OK]
        else:
            icon = JalIcon[JalIcon.CANCEL]
        icon.paint(painter, option.rect, Qt.AlignVCenter | Qt.AlignCenter)
        painter.restore()
        self.paint_grid(painter, option, index)

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.MouseButtonPress:
            if model.data(index, Qt.DisplayRole):  # Toggle value - from 1 to 0 and from 0 to 1
                model.setData(index, 0)
            else:
                model.setData(index, 1)
        return True


# -----------------------------------------------------------------------------------------------------------------------
# This delegate is used to present user with a lookup combobox for selection from predefined constant values
# Constructor parameter 'constant_class' indicates which constant set to be used
# (it should be a descendant of PredefinedList class)
class ConstantLookupDelegate(QStyledItemDelegate):
    def __init__(self, constant_class, parent=None):
        self._parent = parent
        super().__init__(parent=parent)
        self.constants = constant_class()

    def displayText(self, value, locale):
        return self.constants.get_name(value)

    def createEditor(self, aParent, option, index):
        combobox = QComboBox(aParent)
        self.constants.load2combo(combobox)
        return combobox

    def setEditorData(self, editor, index):
        idx = editor.findData(index.model().data(index, Qt.EditRole))
        if idx != -1:
            editor.setCurrentIndex(idx)

    def setModelData(self, editor, model, index):
        model.setData(index, editor.currentData())


# -----------------------------------------------------------------------------------------------------------------------
# Base class for lookup delegate that allows Asset, Peer, Category and Tag selection
class LookupSelectorDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._table = ''
        self._field = ''
        self._selector = None

    def displayText(self, value, locale):
        item_name = JalModel(self, self._table).get_value(self._field, "id", value)
        if item_name is None:
            return ''
        else:
            return item_name

    def createSelector(self, parent) -> None:
        raise NotImplementedError("Method createSelector() isn't defined")

    def createEditor(self, aParent, option, index):
        self.createSelector(aParent)
        assert self._selector is not None, "Selector isn't created in LookupSelectorDelegate descendant"
        return self._selector

    def setEditorData(self, editor: QWidget, index: QModelIndex) -> None:
        editor.selected_id = index.data()

    def setModelData(self, editor, model, index):
        if editor.selected_id:  # Check if lookup index is valid or 0
            model.setData(index, editor.selected_id)
        else:
            model.setData(index, None)  # replace invalid index with NULL value


class CategorySelectorDelegate(LookupSelectorDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._table = "categories"
        self._field = "name"

    def createSelector(self, parent) -> None:
        self._selector = CategorySelector(parent, validate=False)


class TagSelectorDelegate(LookupSelectorDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._table = "tags"
        self._field = "tag"

    def createSelector(self, parent) -> None:
        self._selector = TagSelector(parent, validate=False)


class PeerSelectorDelegate(LookupSelectorDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._table = "agents"
        self._field = "name"

    def createSelector(self, parent) -> None:
        self._selector = PeerSelector(parent, validate=False)


class AssetSelectorDelegate(LookupSelectorDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._table = "assets_ext"
        self._field = "symbol"

    def createSelector(self, parent) -> None:
        self._selector = AssetSelector(parent, validate=False)

# ----------------------------------------------------------------------------------------------------------------------
# This is a helper function for ColoredAmountsDelegate.
# It returns True if given number has move decimal places than Setup.DEFAULT_ACCOUNT_PRECISION
def long_fraction(x: Decimal) -> bool:
    if x is None:
        return False
    try:
        if x.is_nan():
            return False
    except AttributeError:
        return False
    return abs(x - round(x, Setup.DEFAULT_ACCOUNT_PRECISION)) > Decimal('0')

# Display several numbers that provided by the model in form of a list
# Each number is displayed on its own line
# colors - display positive/negative values with green/red color
# signs - display +/- sign before the number if True or only "-" if False
class ColoredAmountsDelegate(QStyledItemDelegate):
    def __init__(self, parent=None, colors=True, signs=True):
        self._view = parent
        self._colors = colors
        self._signs = signs
        super().__init__(parent=parent)

    def paint(self, painter, option, index):
        data = index.model().data(index)
        if not data:
            return
        painter.save()
        color = index.model().data(index, role=Qt.ForegroundRole)
        rect = option.rect
        H = rect.height()
        Y = rect.top()
        rect.setHeight(H / len(data))
        for i, item in enumerate(data):
            rect.moveTop(Y + i * (H / len(data)))
            self.draw_value(option.rect, painter, item, color)
        painter.restore()

    # Displays given value as formatted number with required color (or Green/Red if self._colors is True)
    # If value is None - do nothing, If value is Decimal.NaN - displays Setup.NULL_VALUE
    def draw_value(self, rect, painter, value, color=None):
        text = localize_decimal(value, precision=2, sign=self._signs)
        pen = painter.pen()
        try:
            if self._view.isEnabled():
                if self._colors:
                    if value is not None and not value.is_nan():
                        if value >= 0:
                            pen.setColor(CustomColor.DarkGreen)
                        else:
                            pen.setColor(CustomColor.DarkRed)
                else:
                    if color is not None:
                        pen.setColor(color)
            painter.setPen(pen)
            painter.drawText(rect, Qt.AlignRight | Qt.AlignVCenter, text)
            if long_fraction(value):  # Underline decimal part
                shift = painter.fontMetrics().horizontalAdvance(text[-Setup.DEFAULT_ACCOUNT_PRECISION:])
                painter.drawLine(rect.right() - shift, rect.bottom(), rect.right(), rect.bottom())
        except (TypeError, AttributeError):
            pass



import logging
import xml.etree.ElementTree as xml_tree
from datetime import datetime, timedelta, timezone
from jal.widgets.helpers import timestamp_range, dependency_present
from decimal import Decimal
from io import StringIO, BytesIO

import pandas as pd
from pandas.errors import ParserError
import re
import json
from PySide6.QtCore import Qt, QObject, Signal, Slot, QDate
from PySide6.QtWidgets import QApplication, QDialog, QListWidgetItem

from jal.ui.ui_update_quotes_window import Ui_UpdateQuotesDlg
from jal.constants import MarketDataFeed, PredefinedAsset
from jal.db.asset import JalAsset
from jal.net.web_request import WebRequest
from jal.net.moex import MOEX
try:
    from pypdf import PdfReader
    from pypdf.errors import PdfStreamError
except ImportError:
    pass  # PDF files won't be downloaded without dependency

DATA_SOURCE_ROLE = Qt.UserRole + 1

# ===================================================================================================================
# UI dialog class
# ===================================================================================================================
class QuotesUpdateDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.ui = Ui_UpdateQuotesDlg()
        self.ui.setupUi(self)
        self.ui.StartDateEdit.setDate(QDate.currentDate().addMonths(-1))
        self.ui.EndDateEdit.setDate(QDate.currentDate())
        sources = JalAsset.get_sources_list()
        for source in sources:
            if source != MarketDataFeed.NA:
                item = QListWidgetItem(sources[source], self.ui.SourcesList)
                item.setData(DATA_SOURCE_ROLE, source)
                item.setCheckState(Qt.Checked)
                self.ui.SourcesList.addItem(item)

        # center dialog with respect to parent window
        x = parent.x() + parent.width() / 2 - self.width() / 2
        y = parent.y() + parent.height() / 2 - self.height() / 2
        self.setGeometry(x, y, self.width(), self.height())

    def getStartDate(self):
        return self.ui.StartDateEdit.dateTime().toSecsSinceEpoch()

    def getEndDate(self):
        return self.ui.EndDateEdit.dateTime().toSecsSinceEpoch()

    # Returns a list that contains IDs of all checked data sources
    def getSourceList(self):
        checked = []
        for item_index in range(self.ui.SourcesList.count()):
            item = self.ui.SourcesList.item(item_index)
            if item.checkState() == Qt.Checked:
                checked.append(item.data(DATA_SOURCE_ROLE))
        return checked


# ===================================================================================================================
# Worker class
# ===================================================================================================================
# noinspection SpellCheckingInspection
class QuoteDownloader(QObject):
    download_completed = Signal()
    show_progress = Signal(bool)     # Signal is emitted when downloader wants to start or stop display progress
    update_progress = Signal(float)  # Signal is emitted to report current % of execution

    def __init__(self):
        super().__init__()
        self._request = None
        self._cancelled = False
        self._waiting = False
        self._cbr_codes = None
        self._victoria_quotes = []
        self._victoria_date = None

    @Slot()
    def on_cancel(self):
        self._cancelled = True

    # this method waits for completion of downloading process or for user interrupt (in this case exception is raised)
    def _wait_for_event(self):
        while self._request.isRunning():
            QApplication.processEvents()
            if self._cancelled:
                self._request.quit()
                raise KeyboardInterrupt

    def showQuoteDownloadDialog(self, parent):
        dialog = QuotesUpdateDialog(parent)
        if dialog.exec():
            self.DownloadData(dialog.getStartDate(), dialog.getEndDate(), dialog.getSourceList())
            self.download_completed.emit()

    def DownloadData(self, start_timestamp, end_timestamp, sources_list):
        self._cancelled = False
        self.show_progress.emit(True)
        try:
            if MarketDataFeed.FX in sources_list:
                self.download_currency_rates(start_timestamp, end_timestamp)
            self.download_asset_prices(start_timestamp, end_timestamp, sources_list)
        except KeyboardInterrupt:
            logging.warning(self.tr("Interrupted by user"))
        finally:
            self.show_progress.emit(False)
        if not self._cancelled:
            logging.info(self.tr("Download completed"))

    # Checks for present quotations of 'asset' in given 'currency' and adjusts 'start' timestamp to be at
    # the end of available quotes interval if needed.
    def _adjust_start(self, asset: JalAsset, currency_id: int, start) -> int:
        quotes_begin, quotes_end = asset.quotes_range(currency_id)
        if start < quotes_begin:
            from_timestamp = start
        else:
            from_timestamp = quotes_end if quotes_end > start else start
        return from_timestamp

    def _store_quotations(self, asset: JalAsset, currency_id: int, data: pd.DataFrame) -> None:
        if data is not None:
            data.dropna(inplace=True)
            quotations = []
            for date, quote in data.iterrows():  # Date in pandas dataset is in UTC by default
                quotations.append({'timestamp': int(date.timestamp()), 'quote': quote['Close']})
            asset.set_quotes(quotations, currency_id)

    def download_currency_rates(self, start_timestamp, end_timestamp):
        data_loaders = {
            "RUB": self.CBR_DataReader,
            "EUR": self.ECB_DataReader
        }
        currencies = [x for x in JalAsset.get_currencies() if x.quote_source(None) == MarketDataFeed.FX]
        for base in set([x[1] for x in JalAsset.get_base_currency_history(start_timestamp, end_timestamp)]):
            base_symbol = JalAsset(base).symbol()
            logging.info(self.tr("Loading currency rates for " + base_symbol))
            for i, currency in enumerate([x for x in currencies if x.id() != base]):  # base currency rate is always 1
                if self._cancelled:
                    raise KeyboardInterrupt
                from_timestamp = self._adjust_start(currency, base, start_timestamp)
                try:
                    if from_timestamp <= end_timestamp:
                        data = data_loaders[base_symbol](currency, from_timestamp, end_timestamp)
                        self._store_quotations(currency, base, data)
                except (xml_tree.ParseError, pd.errors.EmptyDataError, KeyError):
                    logging.warning(self.tr("No rates were downloaded for ") + f"{currency.symbol()}/{base_symbol}")
                    continue
                self.update_progress.emit(100.0 * (i+1) / len(currencies))

    def download_asset_prices(self, start_timestamp, end_timestamp, sources_list):
        data_loaders = {
            MarketDataFeed.NA: self.Dummy_DataReader,
            MarketDataFeed.RU: self.MOEX_DataReader,
            MarketDataFeed.EU: self.Euronext_DataReader,
            MarketDataFeed.US: self.Yahoo_Downloader,
            MarketDataFeed.CA: self.TMX_Downloader,
            MarketDataFeed.GB: self.YahooLSE_Downloader,
            MarketDataFeed.FRA: self.YahooFRA_Downloader,
            MarketDataFeed.SMA_VICTORIA: self.Victoria_Downloader,
            MarketDataFeed.COIN: self.Coinbase_Downloader,
            MarketDataFeed.MILAN: self.EuronextMilan_DataReader
        }
        assets = JalAsset.get_active_assets(start_timestamp, end_timestamp)
        assets = [(x['asset'], x['currency']) for x in assets if x['asset'].quote_source(x['currency']) in sources_list]
        logging.info(self.tr("Loading assets prices"))
        for i, (asset, currency) in enumerate(assets):
            from_timestamp = self._adjust_start(asset, currency, start_timestamp)
            try:
                if from_timestamp <= end_timestamp:
                    data = data_loaders[asset.quote_source(currency)](asset, currency, from_timestamp, end_timestamp)
                    self._store_quotations(asset, currency, data)
            except (pd.errors.EmptyDataError, KeyError, json.decoder.JSONDecodeError):
                logging.warning(self.tr("No quotes were downloaded for ") + f"{asset.symbol()}")
                continue
            self.update_progress.emit(100.0 * i / len(assets))

    def PrepareRussianCBReader(self):
        rows = []
        self._request = WebRequest(WebRequest.GET, "http://www.cbr.ru/scripts/XML_valFull.asp")
        self._wait_for_event()
        try:
            xml_root = xml_tree.fromstring(self._request.data())
            for node in xml_root:
                code = node.find("ParentCode").text.strip() if node is not None else None
                iso = node.find("ISO_Char_Code").text if node is not None else None
                rows.append({"ISO_name": iso, "CBR_code": code})
        except xml_tree.ParseError:
            pass
        self._cbr_codes = pd.DataFrame(rows, columns=["ISO_name", "CBR_code"])

    # Empty method to make a unified call for any asset
    def Dummy_DataReader(self, _asset, _currency_id, _start_timestamp, _end_timestamp):
        return None

    def CBR_DataReader(self, currency, start_timestamp, end_timestamp):
        if self._cbr_codes is None:
            self.PrepareRussianCBReader()
        try:
            code = str(self._cbr_codes.loc[self._cbr_codes["ISO_name"] == currency.symbol(), "CBR_code"].values[0]).strip()
        except IndexError:
            logging.debug(self.tr("There are no CBR data for: ") + f"{currency.symbol()}")
            return None
        params = {
            'date_req1': datetime.fromtimestamp(start_timestamp, tz=timezone.utc).strftime('%d/%m/%Y'),
            'date_req2': (datetime.fromtimestamp(end_timestamp, tz=timezone.utc) + timedelta(days=1)).strftime('%d/%m/%Y'), # add 1 day to end_timestamp as CBR sets rate are a day ahead
            'VAL_NM_RQ': code
        }
        url = "http://www.cbr.ru/scripts/XML_dynamic.asp"
        self._request = WebRequest(WebRequest.GET, url, params=params)
        self._wait_for_event()
        xml_root = xml_tree.fromstring(self._request.data())
        rows = []
        for node in xml_root:
            s_date = node.attrib['Date'] if node is not None else None
            s_val = node.find("Value").text if node is not None else None
            s_multiplier = node.find("Nominal").text if node is not None else 1
            rows.append({"Date": s_date, "Close": s_val, "Multiplier": s_multiplier})
        data = pd.DataFrame(rows, columns=["Date", "Close", "Multiplier"])
        data['Date'] = pd.to_datetime(data['Date'], format="%d.%m.%Y", utc=True)
        data['Close'] = [x.replace(',', '.') for x in data['Close']]
        data['Close'] = data['Close'].apply(Decimal)
        data['Multiplier'] = data['Multiplier'].apply(Decimal)
        data['Close'] = data['Close'] / data['Multiplier']
        data.drop('Multiplier', axis=1, inplace=True)
        rates = data.set_index("Date")
        return rates

    def ECB_DataReader(self, currency, start_timestamp, end_timestamp):
        url = f"https://data-api.ecb.europa.eu/service/data/EXR/D.{currency.symbol()}.EUR.SP00.A"
        params = {
            'startPeriod': datetime.fromtimestamp(start_timestamp, tz=timezone.utc).strftime('%Y-%m-%d'),
            'endPeriod': datetime.fromtimestamp(end_timestamp, tz=timezone.utc).strftime('%Y-%m-%d')
        }
        self._request = WebRequest(WebRequest.GET, url, params=params, headers={'Accept': 'text/csv'})
        self._wait_for_event()
        file = StringIO(self._request.data())
        try:
            data = pd.read_csv(file, dtype={'TIME_PERIOD': str, 'OBS_VALUE': str})
        except ParserError:
            return None
        data.rename(columns={'TIME_PERIOD': 'Date', 'OBS_VALUE': 'Close'}, inplace=True)
        data = data[['Date', 'Close']]  # Keep only required columns
        data['Date'] = pd.to_datetime(data['Date'], format="%Y-%m-%d", utc=True)
        data['Close'] = data['Close'].apply(Decimal)   # Convert from str to Decimal
        data['Close'] = Decimal('1') / data['Close']
        data['Close'] = data['Close'].apply(round, args=(10, ))
        rates = data.set_index("Date")
        return rates

    # noinspection PyMethodMayBeStatic
    def MOEX_DataReader(self, asset, currency_id, start_timestamp, end_timestamp, update_symbol=True):
        currency = JalAsset(currency_id).symbol()
        moex_info = MOEX().asset_info(symbol=asset.symbol(currency_id), isin=asset.isin(), currency=currency, special=True)
        if not ('engine' in moex_info and 'market' in moex_info and 'board' in moex_info) or \
                (moex_info['engine'] is None) or (moex_info['market'] is None) or (moex_info['board'] is None):
            logging.warning(f"Failed to find {asset.symbol(currency_id)} ({asset.isin()}) on moex.com")
            return None
        if (moex_info['market'] == 'bonds') and (moex_info['board'] in ['TQCB', 'TQIR']):
            asset_code = asset.isin()   # Corporate bonds are quoted by ISIN
        elif (moex_info['market'] == 'shares') and (moex_info['board'] == 'TQIF'):
            asset_code = asset.isin()   # ETFs are quoted by ISIN
        else:
            asset_code = asset.symbol(currency_id)
        if update_symbol:
            isin = moex_info['isin'] if 'isin' in moex_info else ''
            reg_number = moex_info['reg_number'] if 'reg_number' in moex_info else ''
            expiry = moex_info['expiry'] if 'expiry' in moex_info else 0
            principal = moex_info['principal'] if 'principal' in moex_info else 0
            details = {'isin': isin, 'reg_number': reg_number, 'expiry': expiry, 'principal': principal}
            asset.update_data(details)
        # Get price history
        url = f"http://iss.moex.com/iss/history/engines/{moex_info['engine']}/markets/{moex_info['market']}/" \
              f"boards/{moex_info['board']}/securities/{asset_code}.json"
        params = {
            'from': datetime.fromtimestamp(start_timestamp, tz=timezone.utc).strftime('%Y-%m-%d'),
            'till': datetime.fromtimestamp(end_timestamp, tz=timezone.utc).strftime('%Y-%m-%d')
        }
        self._request = WebRequest(WebRequest.GET, url, params=params)
        self._wait_for_event()
        moex_data = json.loads(self._request.data())
        if 'history' not in moex_data:
            return None
        quotes_data = [dict(zip(moex_data['history']['columns'], x)) for x in moex_data['history']['data']]
        quotes = []
        for x in quotes_data:
            if x['CLOSE']:   # CLOSE field is numeric, it gives a lot of digits after direct conversion to Decimal, so we treat it as a string
                price = Decimal(str(x['CLOSE'])) * Decimal(str(x['FACEVALUE'])) / Decimal('100') if 'FACEVALUE' in x else x['CLOSE']
                quotes.append({"Date": x['TRADEDATE'], "Close": str(price)})
        data = pd.DataFrame(quotes, columns=["Date", "Close"])
        data['Date'] = pd.to_datetime(data['Date'], format="%Y-%m-%d", utc=True)
        data['Close'] = data['Close'].apply(Decimal)
        close = data.set_index("Date")
        return close

    # noinspection PyMethodMayBeStatic
    def Yahoo_Downloader(self, asset, currency_id, start_timestamp, end_timestamp, suffix=''):
        url = f"https://query1.finance.yahoo.com/v8/finance/chart/{asset.symbol(currency_id)+suffix}"
        params = {
            'period1': start_timestamp,
            'period2': end_timestamp,
            'interval': '1d'
        }
        self._request = WebRequest(WebRequest.GET, url, params=params)
        self._wait_for_event()
        web_data = json.loads(self._request.data())
        web_data = web_data['chart']
        if web_data['error'] is not None:
            logging.error(self.tr("Yahoo returned and error: ") + web_data['error'])
            return []
        if len(web_data['result']) != 1:
            logging.error(self.tr("Yahoo returned more then one result: ") + web_data['result'])
            return []
        timestamps = [datetime.fromtimestamp(x, tz=timezone.utc) for x in web_data['result'][0]['timestamp']]
        closes = web_data['result'][0]['indicators']['quote'][0]['close']
        try:
            data = pd.DataFrame({'Date': timestamps, 'Close': closes})
        except ParserError:
            return None
        data['Close'] = data['Close'].apply(Decimal)
        close = data.set_index("Date")
        return close

    # The same as Yahoo_Downloader but it adds ".L" suffix to asset_code and returns prices in GBP
    def YahooLSE_Downloader(self, asset, currency_id, start_timestamp, end_timestamp):
        return self.Yahoo_Downloader(asset, currency_id, start_timestamp, end_timestamp, suffix='.L')

    # The same as Yahoo_Downloader but it adds ".F" suffix to asset_code and returns prices in EUR
    def YahooFRA_Downloader(self, asset, currency_id, start_timestamp, end_timestamp):
        return self.Yahoo_Downloader(asset, currency_id, start_timestamp, end_timestamp, suffix='.F')

    # noinspection PyMethodMayBeStatic
    def Euronext_DataReader(self, asset, currency_id, start_timestamp, end_timestamp):
        suffix = "ETFP" if asset.type() == PredefinedAsset.ETF else "XPAR"  # Dates don't work for ETFP due to glitch on their site
        url = f"https://live.euronext.com/en/ajax/AwlHistoricalPrice/getFullDownloadAjax/{asset.isin()}-{suffix}"
        params = {
            'format': 'csv',
            'decimal_separator': '.',
            'date_form': 'd%2Fm%2FY',
            'op': '',
            'adjusted': 'N',
            'base100': '',
            'startdate': datetime.fromtimestamp(start_timestamp, tz=timezone.utc).strftime('%Y-%m-%d'),
            'enddate': datetime.fromtimestamp(end_timestamp, tz=timezone.utc).strftime('%Y-%m-%d')
        }
        self._request = WebRequest(WebRequest.GET, url, params=params)
        self._wait_for_event()
        quotes = self._request.data()
        quotes_text = quotes.replace(u'\ufeff', '').splitlines()    # Remove BOM from the beginning
        if len(quotes_text) < 4:
            logging.warning(self.tr("Euronext quotes history reply is too short: ") + quotes)
            return None
        if quotes_text[0] != '"Historical Data"':
            logging.warning(self.tr("Euronext quotes header not found in: ") + quotes)
            return None
        if quotes_text[2] != asset.isin():
            logging.warning(self.tr("Euronext quotes ISIN mismatch in: ") + quotes)
            return None
        quotes_text = [x.replace("'", "") for x in quotes_text]   # Some lines have occasional quote in some places
        quotes = "\n".join(quotes_text)
        file = StringIO(quotes)
        try:
            data = pd.read_csv(file, header=3, sep=';', dtype={'Date': str, 'Close': str}, index_col=False)
        except ParserError:
            return None
        data['Date'] = pd.to_datetime(data['Date'], format="%d/%m/%Y", utc=True)
        data = data[data.Date >= datetime.fromtimestamp(start_timestamp, tz=timezone.utc)]   # There is a bug on Euronext side - it returns full set regardless of date
        data = data[data.Date <= datetime.fromtimestamp(end_timestamp, tz=timezone.utc)]
        data['Close'] = data['Close'].apply(Decimal)
        data = data.drop(columns=['Open', 'High', 'Low', 'Last', 'Number of Shares', 'Number of Trades', 'Turnover', 'vwap'],
                         errors='ignore')  # Ignore errors as some columns might be missing
        close = data.set_index("Date")
        close.sort_index(inplace=True)
        return close

    # noinspection PyMethodMayBeStatic
    def EuronextMilan_DataReader(self, asset, currency_id, start_timestamp, end_timestamp):
        suffix = "ETF" if asset.type() == PredefinedAsset.ETF else "MTA"
        url = "https://charts.borsaitaliana.it/charts/services/ChartWService.asmx/GetPricesWithVolume"
        params = {
            "request": {
                "SampleTime":"1d",
                "TimeFrame":None,
                "RequestedDataSetType":"ohlc",
                "ChartPriceType":"price",
                "Key": asset.symbol(currency_id) + "." + suffix,
                "OffSet":0,
                "FromDate":start_timestamp,
                "ToDate":end_timestamp,
                "UseDelay":False,
                "KeyType":"Topic",
                "KeyType2":"Topic",
                "Language":"en-US"
            }
        }
        self._request = WebRequest(WebRequest.POST_JSON, url, params=params, headers={'Accept': 'text/csv'})
        self._wait_for_event()
        json_content = json.loads(self._request.data())
        if 'd' not in json_content:
            return None
        data = pd.DataFrame(json_content['d'], columns=["Date", "Unk", "O", "H", "L", "Close", "V"])
        data = data.drop(columns=["Unk", "O", "H", "L", "V"], errors='ignore')
        data['Date'] = pd.to_datetime(data['Date'], unit='ms', utc=True)
        data['Close'] = data['Close'].astype('str').apply(Decimal)
        close = data.set_index("Date")
        close.sort_index(inplace=True)
        return close

    # noinspection PyMethodMayBeStatic
    def TMX_Downloader(self, asset, _currency_id, start_timestamp, end_timestamp):
        url = 'https://app-money.tmx.com/graphql'
        params = {
            "operationName": "getCompanyPriceHistoryForDownload",
            "variables":
                {
                    "symbol": asset.symbol(),
                    "start": datetime.fromtimestamp(start_timestamp, tz=timezone.utc).strftime('%Y-%m-%d'),
                    "end": datetime.fromtimestamp(end_timestamp, tz=timezone.utc).strftime('%Y-%m-%d'),
                    "adjusted": False,
                    "adjustmentType": None,
                    "unadjusted": True
                },
            "query": "query getCompanyPriceHistoryForDownload($symbol: String!, $start: String, $end: String, $adjusted: Boolean, $adjustmentType: String, $unadjusted: Boolean) "
                     "{getCompanyPriceHistoryForDownload(symbol: $symbol, start: $start, end: $end, adjusted: $adjusted, adjustmentType: $adjustmentType, unadjusted: $unadjusted) "
                     "{ datetime closePrice}}"
        }
        self._request = WebRequest(WebRequest.POST_JSON, url, params=params)
        self._wait_for_event()
        json_content = json.loads(self._request.data())
        result_data = json_content['data'] if 'data' in json_content else None
        if 'getCompanyPriceHistoryForDownload' in result_data:
            price_array = result_data['getCompanyPriceHistoryForDownload']
        else:
            logging.warning(self.tr("Can't parse data for TSX quotes: ") + json_content)
            return None
        data = pd.DataFrame(price_array)
        data.rename(columns={'datetime': 'Date', 'closePrice': 'Close'}, inplace=True)
        data['Date'] = pd.to_datetime(data['Date'], format="%Y-%m-%d", utc=True)
        data['Close'] = data['Close'].apply(str)   # Convert from float to str
        data['Close'] = data['Close'].apply(Decimal)   # Convert from str to Decimal
        close = data.set_index("Date")
        close.sort_index(inplace=True)
        return close

    def Victoria_Downloader(self, asset, _currency_id, _start_timestamp, _end_timestamp):
        quotes = self._get_victoria_quotes()
        if not quotes:
            return None
        # Filter asset price
        asset_quotes = [x for x in quotes if x['name'] == asset.name()]
        if len(asset_quotes) != 1:
            logging.warning(self.tr("Can't find quote for Victoria Seguros fund: ") + asset.name())
            return None
        data = pd.DataFrame([{'Date': self._victoria_date, 'Close': asset_quotes[0]['price']}])
        close = data.set_index("Date")
        return close

    def _get_victoria_quotes(self) -> list:
        months = ['janeiro', 'fevereiro', 'março', 'abril', 'maio', 'junho', 'julho', 'agosto', 'setembro', 'outubro',
                  'novembro', 'dezembro']
        if self._victoria_quotes:
            return self._victoria_quotes
        if not dependency_present(['pypdf']):
            logging.warning(self.tr("Package pypdf not found for PDF parsing."))
            return []
        # Download PDF-file with current quotes
        self._request = WebRequest(WebRequest.GET, "https://www.victoria-seguros.pt/cdu-services/UNIDADES_PARTICIPACAO_VIDA_OPC1_UP", binary=True)
        self._wait_for_event()
        pdf_data = self._request.data()
        if not pdf_data:
            return []
        try:
            pdf = PdfReader(BytesIO(pdf_data))
        except PdfStreamError:
            logging.error(self.tr("Can't parse server response as pdf: ") + str(pdf_data))
            return []
        if len(pdf.pages) != 1:
            logging.warning(self.tr("Unexpected number of pages in Victoria Seguros document: ") + len(pdf.pages))
            return []
        # Get text from downloaded PDF-file together with font type and font size of each text fragment
        parts = []
        def visitor_body(text, cm, tm, font_dict, font_size):
            font = {} if font_dict is None else font_dict
            if text:
                parts.append({'size': font_size, 'font': font['/BaseFont'], 'text': text})
        pdf.pages[0].extract_text(visitor_text=visitor_body)
        # Combine text fragments into lines according to line breaks and font changes
        lines = []
        line = ''
        font = ''
        for part in parts:
            if part['text'] == '\n':
                lines.append(line)
                line = ''
                continue
            if part['font'] != font:
                font = part['font']
                if line:
                    lines.append(line)
                line = part['text']
                continue
            line += part['text']
        # Get quote date
        match = re.match(r"(\d\d?) de (\w+) de (\d\d\d\d)", lines[-1])
        if match is None:
            logging.warning(self.tr("Can't parse date from Victoria Seguros file"))
            return []
        day, month, year = match.groups()
        self._victoria_date = datetime(int(year), months.index(month) + 1, int(day))
        # Get quotation lines from the file
        self._victoria_quotes = []
        for line in lines:
            match = re.match(r"(.*) EUR (\d+[,.]\d+)", line)
            if match is None:
                continue
            fund_name, price = match.groups()
            self._victoria_quotes.append({'name': fund_name, 'price': Decimal(price.replace(',', '.'))})
        return self._victoria_quotes

    def Coinbase_Downloader(self, asset, currency_id, start_timestamp, end_timestamp):
        currency_symbol = JalAsset(currency_id).symbol()
        url = f"https://api.coinbase.com/v2/prices/{asset.symbol()}-{currency_symbol}/spot"
        quotes = []
        for ts in timestamp_range(start_timestamp, end_timestamp):
            date_string = datetime.fromtimestamp(ts, tz=timezone.utc).strftime('%Y-%m-%d')
            self._request = WebRequest(WebRequest.GET, url, params={'date': date_string})
            self._wait_for_event()
            try:
                result_data = json.loads(self._request.data())
                quote = result_data['data']['amount']
            except (json.decoder.JSONDecodeError, KeyError):
                continue
            quotes.append({"Date": datetime.fromtimestamp(ts, tz=timezone.utc), "Close": quote})
        data = pd.DataFrame(quotes, columns=["Date", "Close"])
        data['Close'] = data['Close'].apply(Decimal)
        close = data.set_index("Date")
        return close

    # Returns a list of currencies supported by Coinbase exchange as a list of {'symbol', 'name'}
    @staticmethod
    def Coinbase_GetCurrencyList() -> list:
        request = WebRequest(WebRequest.GET, "https://api.coinbase.com/v2/currencies/crypto")
        while request.isRunning():
            QApplication.processEvents()
        result_data = json.loads(request.data())
        data = result_data['data']
        assets = [{'symbol': x['code'], 'name': x['name']} for x in data if x['type'] == 'crypto']
        return assets



from decimal import Decimal
from jal.constants import PredefinedAsset, PredefinedCategory
from jal.db.operations import AssetPayment, CorporateAction
from jal.data_export.taxes import TaxReport
from jal.db.category import JalCategory


class TaxesPortugal(TaxReport):
    currency_name = 'EUR'
    country_name = 'portugal'

    def __init__(self):
        super().__init__()
        self._processed_trade_qty = {}  # It will handle {trade_id: qty} records to keep track of already processed qty
        self.reports = {
            "Dividends": (self.prepare_dividends, "tax_prt_dividends.json"),
            "Shares": (self.prepare_stocks_and_etf, "tax_prt_shares.json"),
            "Interest": (self.prepare_broker_interest, "tax_prt_interests.json")
        }

    def prepare_dividends(self):
        dividends_report = []
        dividends = self.dividends_list()
        dividends = [x for x in dividends if x.amount(self.account_currency.id()) > 0]  # Skip negative dividends
        for dividend in dividends:
            country = dividend.asset().country()
            note = ''
            if dividend.subtype() == AssetPayment.StockDividend:
                note = "Stock dividend"
            if dividend.subtype() == AssetPayment.StockVesting:
                note = "Stock vesting"
            line = {
                'report_template': "dividend",
                'payment_date': dividend.timestamp(),
                'symbol': dividend.asset().symbol(self.account_currency.id()),
                'full_name': dividend.asset().name(),
                'isin': dividend.asset().isin(),
                'amount': dividend.amount(self.account_currency.id()),
                'tax': dividend.tax(),
                'rate': self.account_currency.quote(dividend.timestamp(), self._currency_id)[1],
                'country': country.name(language='en'),
                'tax_treaty': "Y" if self.has_tax_treaty_with(country.code()) else "N",
                'amount_eur': round(dividend.amount(self._currency_id), 2),
                'tax_eur': round(dividend.tax(self._currency_id), 2),
                'note': note
            }
            dividends_report.append(line)
        self.insert_totals(dividends_report, ["amount", "amount_eur", "tax", "tax_eur"])
        return dividends_report

# ----------------------------------------------------------------------------------------------------------------------
    def prepare_stocks_and_etf(self):
        deals_report = []
        ns = not self.use_settlement
        trades = self.trades_list([PredefinedAsset.Stock, PredefinedAsset.ETF])
        for trade in trades:
            note = ''
            if ns:
                rate_ts = trade.close_operation().timestamp() if self.one_currency_rate else 0
            else:
                rate_ts = trade.close_operation().settlement() if self.one_currency_rate else 0
            if trade.qty() >= Decimal('0'):  # Long trade
                value_realization = round(trade.close_amount(no_settlement=ns), 2)
                value_realization_eur = round(trade.close_amount(self._currency_id, rate_ts=rate_ts, no_settlement=ns), 2)
                value_acquisition = round(trade.open_amount(no_settlement=ns), 2) + round(trade.fee(), 2)
                value_acquisition_eur = round(trade.open_amount(self._currency_id, rate_ts=rate_ts, no_settlement=ns), 2) + round(trade.fee(self._currency_id, rate_ts=rate_ts), 2)
            else:  # Short trade
                value_realization = round(trade.open_amount(no_settlement=ns), 2)
                value_realization_eur = round(trade.open_amount(self._currency_id, rate_ts=rate_ts, no_settlement=ns), 2)
                value_acquisition = round(trade.close_amount(no_settlement=ns), 2) + round(trade.fee(), 2)
                value_acquisition_eur = round(trade.close_amount(self._currency_id, rate_ts=rate_ts, no_settlement=ns), 2) + round(trade.fee(self._currency_id, rate_ts=rate_ts), 2)
            for modifier in trade.modified_by():
                note = note + modifier.description() + "\n"
            line = {
                'report_template': "trade",
                'symbol': trade.asset().symbol(self.account_currency.id()),
                'isin': trade.asset().isin(),
                'qty': trade.qty(),
                'o_type': "Buy" if trade.qty() >= Decimal('0') else "Sell",
                'o_number': trade.open_operation().number(),
                'o_date': trade.open_operation().timestamp(),
                'os_date': trade.open_operation().settlement(),
                'o_price': trade.open_price(),
                'o_fee': trade.open_fee(),
                'o_fee_eur': trade.open_fee(self._currency_id, rate_ts=rate_ts),
                'o_amount': round(trade.open_amount(no_settlement=ns), 2),
                'o_amount_eur': round(trade.open_amount(self._currency_id, rate_ts=rate_ts, no_settlement=ns), 2),
                'c_type': "Sell" if trade.qty() >= Decimal('0') else "Buy",
                'c_number': trade.close_operation().number(),
                'c_date': trade.close_operation().timestamp(),
                'cs_date': trade.close_operation().settlement(),
                'c_price': trade.close_operation().price(),
                'c_fee': trade.close_fee(),
                'c_fee_eur': trade.close_fee(self._currency_id, rate_ts=rate_ts),
                'c_amount': round(trade.close_amount(no_settlement=ns), 2),
                'c_amount_eur': round(trade.close_amount(self._currency_id, rate_ts=rate_ts, no_settlement=ns), 2),
                'profit': value_realization - value_acquisition,
                'profit_eur': value_realization_eur - value_acquisition_eur,
                'rate': self.account_currency.quote(rate_ts, self._currency_id)[1],
                'note': note
            }
            deals_report.append(line)
        self.insert_totals(deals_report, ["income_eur", "spending_eur", "profit_eur", "profit"])
        return deals_report

# ----------------------------------------------------------------------------------------------------------------------
    def prepare_broker_interest(self):
        interests_report = []
        interest_operations = JalCategory(PredefinedCategory.Interest).get_operations(self.year_begin, self.year_end)
        interest_operations = [x for x in interest_operations if x.account_id() == self.account.id()]
        for operation in interest_operations:
            rate = self.account_currency.quote(operation.timestamp(), self._currency_id)[1]
            interests = [x for x in operation.lines() if x['category_id'] == PredefinedCategory.Interest]
            for interest in interests:
                amount = Decimal(interest['amount'])
                line = {
                    'report_template': "interest",
                    'payment_date': operation.timestamp(),
                    'rate': rate,
                    'amount': amount,
                    'amount_eur': round(amount * rate, 2),
                    'note': interest['note']
                }
                interests_report.append(line)
        # Process cash payments out of corporate actions
        payments = CorporateAction.get_payments(self.account)
        payments = [x for x in payments if self.year_begin <= x['timestamp'] <= self.year_end]
        for payment in payments:
            rate = self.account_currency.quote(payment['timestamp'], self._currency_id)[1]
            line = {
                'report_template': "interest",
                'payment_date': payment['timestamp'],
                'rate': rate,
                'amount': payment['amount'],
                'amount_eur': round(payment['amount'] * rate, 2),
                'note': payment['note']
            }
            interests_report.append(line)
        self.insert_totals(interests_report, ["amount", "amount_eur"])
        return interests_report



# Base classes for Category, Tag and Peer reports
from decimal import Decimal
from PySide6.QtCore import Qt, QAbstractTableModel
from PySide6.QtGui import QFontDatabase
from PySide6.QtWidgets import QHeaderView
from jal.db.asset import JalAsset
from jal.db.account import JalAccount
from jal.db.category import JalCategory
from jal.db.tag import JalTag
from jal.db.peer import JalPeer
from jal.db.helpers import localize_decimal
from jal.db.operations import LedgerTransaction
from jal.widgets.helpers import ts2dt
from jal.widgets.delegates import ColoredAmountsDelegate


class ReportOperationsModel(QAbstractTableModel):
    # hidden_column indicates which column to hide from display in the view
    def __init__(self, parent_view, hidden_column=None):
        super().__init__(parent_view)
        self._columns = [self.tr("Timestamp"), self.tr("Account"), self.tr("Peer"), self.tr("Category"), self.tr("Tag"),
                         self.tr("Notes"), self.tr("Amount"), self.tr("Currency")]
        self._view = parent_view
        self._hidden = hidden_column
        self._amount_delegate = None
        self._data = []
        self._begin = 0
        self._end = 0
        self._total = Decimal('0')
        self._total_currency = 0
        self._total_currency_name = ''
        self._bold_font = QFontDatabase.systemFont(QFontDatabase.GeneralFont)
        self._bold_font.setBold(True)
        self.prepareData()

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def get_operation(self, row):
        if (row >= 0) and (row < self.rowCount()):
            return self._data[row]['otype'], self._data[row]['oid']
        else:
            return [0, 0]

    def data(self, index, role=Qt.DisplayRole, field=''):
        row = index.row()
        odata = self._data[row]
        if not index.isValid():
            return None
        try:
            operation = LedgerTransaction().get_operation(odata['otype'], odata['oid'], odata['opart'])
        except IndexError as e:
            if str(e) == LedgerTransaction.NoOpException:
                return None
            raise e
        if role == Qt.DisplayRole:
            return self.data_text(operation, index.column(), odata)
        if role == Qt.DecorationRole and index.column() == 0:
            return operation.icon()

    def data_text(self, operation, column, odata):
        if column == 0:
            return ts2dt(operation.timestamp())
        elif column == 1:
            return operation.account_name()
        elif column == 2:
            return JalPeer(operation.peer()).name()
        elif column == 3:
            return JalCategory(odata['category_id']).name()
        elif column == 4:
            return JalTag(odata['tag_id']).name()
        elif column == 5:
            return operation.description(part_only=True)
        elif column == 6:
            return operation.value_change(part_only=True)
        elif column == 7:
            return operation.value_currency()
        else:
            assert False, "Unexpected column number"

    # Is used by view to display footer Title, Total amount and Total currency with right font and alignment in columns 3-5
    def footerData(self, section: int, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section == 6:
                return localize_decimal(self._total, precision=2)
            elif section == 7:
                return self._total_currency_name
        elif role == Qt.FontRole:
            return self._bold_font
        elif role == Qt.TextAlignmentRole:
            if section == 6:
                return Qt.AlignRight | Qt.AlignVCenter
            return Qt.AlignLeft | Qt.AlignVCenter
        return None

    def configureView(self):
        self._view.setColumnWidth(0, self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.2)
        self._view.setColumnWidth(1, 150)
        self._view.setColumnWidth(2, 150)
        self._view.setColumnWidth(3, 150)
        self._view.setColumnWidth(4, 150)
        self._view.horizontalHeader().setSectionResizeMode(5, QHeaderView.Stretch)
        self._view.horizontalHeader().setFont(self._bold_font)
        self._amount_delegate = ColoredAmountsDelegate(self._view)
        self._view.setItemDelegateForColumn(6, self._amount_delegate)
        self._view.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)  # row size is adjusted in data() method
        if self._hidden is not None:
            self._view.setColumnHidden(self._hidden, True)

    # Triggers view update if display parameters were changed
    def updateView(self, update: bool, dates_range: tuple, total_currency_id: int):
        if self._begin != dates_range[0]:
            self._begin = dates_range[0]
            update = True
        if self._end != dates_range[1]:
            self._end = dates_range[1]
            update = True
        if self._total_currency != total_currency_id:
            self._total_currency = total_currency_id
            self._total_currency_name = JalAsset(total_currency_id).symbol()
            update = True
        if update:
            self.prepareData()
            self.configureView()

    # self._data array should be populated by child classes.
    # Then they call this method to calculate total value.
    def prepareData(self):
        self.beginResetModel()
        self._total = Decimal('0')
        for line in self._data:
            account_currency = JalAccount(line['account_id']).currency()
            amount = -Decimal(line['amount'])
            if account_currency == self._total_currency:
                self._total += amount
            else:
                self._total += amount * JalAsset(account_currency).quote(line['timestamp'], self._total_currency)[1]
        self.endResetModel()



from decimal import Decimal
from jal.db.db import JalDB
from jal.db.helpers import format_decimal
import jal.db.account
import jal.db.operations
from jal.db.asset import JalAsset


# ----------------------------------------------------------------------------------------------------------------------
# Class that represents open trade and provides some methods equal to JalClosedTrades to make compatible calls
class JalOpenTrade(JalDB):
    def __init__(self, operation, price, qty, adjustments=(Decimal('1'), Decimal('1'))) -> None:
        super().__init__()
        self._op = operation
        self._price = price
        self._qty = qty
        self._adj_price = adjustments[0]   # Historical adjustments of price and quantity that happened
        self._adj_qty = adjustments[1]     # during holding of this position

    def open_operation(self):
        return self._op

    def open_price(self, adjusted=False) -> Decimal:
        if adjusted:
            return self._price * self._adj_price
        else:
            return self._price

    def open_qty(self, adjusted=False) -> Decimal:
        if adjusted:
            return self._qty * self._adj_qty
        else:
            return self._qty

    # Method to adjust price of open position
    def set_price(self, new_price: Decimal):
        self._price = new_price

    # Method to adjust size of open position
    def set_qty(self, new_qty: Decimal):
        self._qty = new_qty

    def p_adjustment(self) -> Decimal:
        return self._adj_price

    def q_adjustment(self) -> Decimal:
        return self._adj_qty


# ----------------------------------------------------------------------------------------------------------------------
class JalClosedTrade(JalDB):
    def __init__(self, id: int = 0) -> None:
        super().__init__()
        self._id = id
        self._data = self._read("SELECT account_id, asset_id, open_otype, open_oid, open_timestamp, open_price, "
                                "open_qty, close_otype, close_oid, close_timestamp, close_price, close_qty, c_price, c_qty "
                                "FROM trades_closed WHERE id=:id", [(":id", self._id)], named=True)
        if self._data:
            self._account = jal.db.account.JalAccount(self._data['account_id'])
            self._asset = JalAsset(self._data['asset_id'])
            self._open_op = jal.db.operations.LedgerTransaction.get_operation(self._data['open_otype'], self._data['open_oid'], jal.db.operations.Transfer.Incoming)
            self._close_op = jal.db.operations.LedgerTransaction.get_operation(self._data['close_otype'], self._data['close_oid'], jal.db.operations.Transfer.Outgoing)
            self._open_price = Decimal(self._data['open_price'])
            self._open_qty = Decimal(self._data['open_qty'])
            self._close_price = Decimal(self._data['close_price'])
            self._qty = Decimal(self._data['close_qty'])
            self._adj_price = Decimal(self._data['c_price'])
            self._adj_qty = Decimal(self._data['c_qty'])
        else:
            self._account = self._asset = self._open_op = self._close_op = None
            self._open_price = self._open_qty = self._close_price = self._qty = Decimal('0')
            self._adj_price = self._adj_qty = Decimal('1')

    @classmethod
    # Creates a record in 'trades_closed' table based on given data:
    # o_trade - trade JalClosedTrade that initiates the trade (may also be AssetPayment as Stock Dividend or Vesting)
    # c_operation - LedgerTransaction that closed current trade
    # o_qty and c_qty - how many items were used from opening and closing trades respectively
    def create_from_trades(cls, o_trade: JalOpenTrade, c_operation, qty):
        o_operation = o_trade.open_operation()
        o_qty = qty / o_trade.q_adjustment()   # Calculate adjusted amount for opening trade
        c_price = o_trade.open_price() if c_operation.price() is None else c_operation.price()
        _ = cls._exec(
            "INSERT INTO trades_closed(account_id, asset_id, open_otype, open_oid, open_timestamp, open_price, "
            "open_qty, close_otype, close_oid, close_timestamp, close_price, close_qty, c_price, c_qty) "
            "VALUES(:account_id, :asset_id, :open_otype, :open_oid, :open_timestamp, :open_price, :open_qty, "
            ":close_otype, :close_oid, :close_timestamp, :close_price, :close_qty, :c_price, :c_qty)",
            [(":account_id", c_operation.account().id()), (":asset_id", c_operation.asset().id()),
             (":open_otype", o_operation.type()), (":open_oid", o_operation.id()),
             (":open_timestamp", o_operation.timestamp()), (":open_price", format_decimal(o_trade.open_price())),
             (":open_qty", format_decimal(o_qty)), (":close_otype", c_operation.type()),
             (":close_oid", c_operation.id()), (":close_timestamp", c_operation.timestamp()),
             (":close_price", format_decimal(c_price)), (":close_qty", format_decimal(qty)),
             (":c_price", format_decimal(o_trade.p_adjustment())), (":c_qty", format_decimal(o_trade.q_adjustment()))])

    def dump(self) -> list:
        return [
            self._asset.symbol(self._account.currency()),
            self._open_op.timestamp(),
            self._close_op.timestamp(),
            self._open_price,
            self._close_price,
            self._qty,
            self.fee(),
            self.profit(),
            self.profit(percent=True)
        ]

    def id(self) -> int:
        return self._id

    def asset(self) -> JalAsset:
        return self._asset

    def symbol(self) -> str:
        return self._asset.symbol(self._account.currency())

    def open_operation(self):
        return self._open_op

    def close_operation(self):
        return self._close_op

    def open_qty(self, adjusted=False) -> Decimal:
        if adjusted:
            return self._open_qty * self._adj_qty
        else:
            return self._open_qty

    def qty(self) -> Decimal:
        return self._qty

    def open_timestamp(self) -> int:
        return self._open_op.timestamp()

    def close_timestamp(self) -> int:
        return self._close_op.timestamp()

    def open_price(self, adjusted=False) -> Decimal:
        if adjusted:
            return self._open_price * self._adj_price
        else:
            return self._open_price

    def close_price(self) -> Decimal:
        return self._close_price

    def p_adjustment(self) -> Decimal:
        return self._adj_price

    def q_adjustment(self) -> Decimal:
        return self._adj_qty

    def cost_basis(self) -> Decimal:
        return self._adj_price * self._adj_qty

    # If currency_id is different from trade account currency then adjusts value to the rate of currency for
    # given timestamp and returns it. Otherwise, simply returns unchanged value
    def adjusted(self, value: Decimal, currency_id: int, timestamp: int) -> Decimal:
        if currency_id and currency_id != self._account.currency():
            return value * JalAsset(self._account.currency()).quote(timestamp, currency_id)[1]
        else:
            return value

    # Returns opening amount of the trade (open price x qty)
    # currency_id != 0 - convert amount into given currency (using settlement date rate by default)
    # rate_ts - defines timestamp of currency rate if not zero
    # no_settlement - transaction date is used for currency conversion, not settlement date
    # full==True - provide full amount of opening trade (without correction due to corporate actions)
    def open_amount(self, currency_id: int = 0, rate_ts: int = 0, no_settlement=False, full=False) -> Decimal:
        cost_basis = Decimal('1') if full else self.cost_basis()
        timestamp = self._open_op.timestamp() if no_settlement else self._open_op.settlement()
        timestamp = rate_ts if rate_ts else timestamp
        return self.adjusted(cost_basis * self.open_price() * abs(self.open_qty()), currency_id, timestamp)

    # Returns closing amount of the trade (close price x qty)
    # currency_id != 0 - convert amount into given currency (using settlement date rate by default)
    # rate_ts - defines timestamp of currency rate if not zero
    # no_settlement - transaction date is used for currency conversion, not settlement date
    def close_amount(self, currency_id: int = 0, rate_ts: int = 0, no_settlement=False) -> Decimal:
        timestamp = self._close_op.timestamp() if no_settlement else self._close_op.settlement()
        timestamp = rate_ts if rate_ts else timestamp
        return self.adjusted(self._close_price * abs(self._qty), currency_id, timestamp)

    # Fee of opening part of the deal
    # if currency_id isn't 0 then returns fee converted into given currency
    # If full==True, then provide full fee of opening trade (without correction due to corporate actions)
    # rate_ts - defines timestamp of currency rate if not zero
    def open_fee(self, currency_id: int = 0, rate_ts: int = 0, full: bool = False) -> Decimal:
        cost_basis = Decimal('1') if full else self.cost_basis()
        if self._open_op.type() == jal.db.operations.LedgerTransaction.Trade:
            o_fee = cost_basis * self._open_op.fee() * abs(self._qty / (self._open_op.qty() * self.q_adjustment()))
            timestamp = rate_ts if rate_ts else self._open_op.timestamp()
            return self.adjusted(o_fee, currency_id, timestamp)
        else:
            return Decimal('0')

    # Fee of closing part of the deal
    # if currency_id isn't 0 then returns fee converted into given currency
    # rate_ts - defines timestamp of currency rate if not zero
    def close_fee(self, currency_id: int = 0, rate_ts: int = 0) -> Decimal:
        if self._close_op.type() == jal.db.operations.LedgerTransaction.Trade:
            c_fee = self._close_op.fee() * abs(self._qty / self._close_op.qty())
            timestamp = rate_ts if rate_ts else self._close_op.timestamp()
            return self.adjusted(c_fee, currency_id, timestamp)
        else:
            return Decimal('0')

    # Total fee for the trade
    def fee(self, currency_id: int = 0, rate_ts: int = 0) -> Decimal:
        return self.open_fee(currency_id, rate_ts) + self.close_fee(currency_id, rate_ts)

    def profit(self, percent=False) -> Decimal:
        profit = self._qty * self._close_price - self.open_price(adjusted=True) * self.open_qty(adjusted=True) - self.fee()
        if percent:
            profit = Decimal('100') * profit / (self.open_price(adjusted=True) * self.open_qty(adjusted=True)) if self._open_price else Decimal('0')
        return profit

    # Returns a list of LedgerTransactions that modified position after its opening
    # (i.e. transfers or corporate actions that happened during position lifetime)
    def modified_by(self) -> list:
        query = self._exec("SELECT DISTINCT m_otype, m_oid FROM trades_opened "
                           "WHERE otype=:otype AND oid=:oid AND m_otype!=:trade",
                           [(":otype", self._open_op.type()), (":oid", self._open_op.id()),
                            (":trade", jal.db.operations.LedgerTransaction.Trade)])
        modifiers = []
        while query.next():
            otype, oid = self._read_record(query)
            modifiers.append(jal.db.operations.LedgerTransaction.get_operation(otype, oid, jal.db.operations.Transfer.Outgoing))
        return modifiers



from PySide6.QtCore import Qt, Slot, QDate, QAbstractTableModel, QModelIndex
from PySide6.QtGui import QFontDatabase
from PySide6.QtWidgets import QHeaderView
from jal.constants import CustomColor
from jal.db.ledger import Ledger
from jal.db.helpers import localize_decimal
from jal.db.operations import LedgerTransaction
from jal.widgets.helpers import ts2dt
from jal.widgets.delegates import ColoredAmountsDelegate, long_fraction


#-----------------------------------------------------------------------------------------------------------------------

class OperationsModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [self.tr("Timestamp"), self.tr("Account"), self.tr("Notes"),
                         self.tr("Amount"), self.tr("Balance"), self.tr("Currency")]
        self._view = parent_view
        self._amount_delegate = None
        self._total_delegate = None
        self._data = []
        self._begin = 0
        self._end = 0
        self._account = 0
        self._bold_font = QFontDatabase.systemFont(QFontDatabase.GeneralFont)
        self._bold_font.setBold(True)
        self.prepareData()

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def get_operation(self, row):
        if (row >= 0) and (row < self.rowCount()):
            return self._data[row]['otype'], self._data[row]['oid']
        else:
            return [0, 0]

    def data(self, index, role=Qt.DisplayRole, field=''):
        row = index.row()
        odata = self._data[row]
        if not index.isValid():
            return None
        try:
            operation = LedgerTransaction().get_operation(odata['otype'], odata['oid'], odata['opart'])
        except IndexError as e:
            if str(e) == LedgerTransaction.NoOpException:
                return None
            raise e
        if role == Qt.DisplayRole:
            return self.data_text(operation, index.column())
        if role == Qt.DecorationRole and index.column() == 0:
            return operation.icon()
        if role == Qt.FontRole and index.column() == 0:
            # below line isn't related with font, it is put here to be called for each row minimal times (ideally 1)
            self._view.setRowHeight(row, self._view.verticalHeader().fontMetrics().height() * operation.view_rows())
            return self._view.font()
        if role == Qt.ForegroundRole and self._view.isEnabled():
            if index.column() == 4 and operation.reconciled():
                return CustomColor.Blue
        if role == Qt.ToolTipRole:
            if index.column() == 0:
                return operation.name()
            elif index.column() == 3 or index.column() == 4:
                data = self.data_text(operation, index.column())
                if any([long_fraction(x) for x in data]):
                    return '\n'.join([localize_decimal(x) for x in data])
        if role == Qt.TextAlignmentRole:
            if index.column() == 3 or index.column() == 4:
                return int(Qt.AlignRight)
            return int(Qt.AlignLeft)
        if role == Qt.UserRole:  # return underlying data for given field extra parameter
            return odata[field]

    def data_text(self, operation, column):
        if column == 0:
            date_time = ts2dt(operation.timestamp())
            if operation.number() and operation.type() != LedgerTransaction.Transfer:  # Transfer is 1-liner
                date_time += f"\n# {operation.number()}"
            return date_time
        elif column == 1:
            if operation.asset_name() and operation.type() != LedgerTransaction.Transfer:
                return operation.account_name() + "\n" + operation.asset_name()
            else:
                return operation.account_name()
        elif column == 2:
            return operation.description()
        elif column == 3:
            return operation.value_change()
        elif column == 4:
            return operation.value_total()
        elif column == 5:
            return operation.value_currency()
        else:
            assert False, "Unexpected column number"

    def configureView(self):
        self._view.setColumnWidth(0, self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.2)
        self._view.setColumnWidth(1, 300)
        self._view.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)
        self._view.horizontalHeader().setFont(self._bold_font)
        self._amount_delegate = ColoredAmountsDelegate(self._view)
        self._total_delegate = ColoredAmountsDelegate(self._view, colors=False, signs=False)
        self._view.setItemDelegateForColumn(3, self._amount_delegate)
        self._view.setItemDelegateForColumn(4, self._total_delegate)
        self._view.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)  # row size is adjusted in data() method

    @Slot()
    def setAccount(self, account_id):
        if self._account != account_id:
            self._account = account_id
            self.prepareData()

    def getAccount(self):
        return self._account

    @Slot()
    def setDateRange(self, start, end=0):
        self._begin = start
        if end:
            self._end = end
        else:
            self._end = QDate.currentDate().endOfDay(Qt.UTC).toSecsSinceEpoch()
        self.prepareData()

    def update(self):
        self.prepareData()

    @Slot()
    def refresh(self):
        idx = self._view.selectionModel().selection().indexes()
        idx = idx[0] if idx else QModelIndex()  # Take first selected or empty index
        if self._view.model() != self:          # View uses some proxy model
            idx = self._view.model().mapToSource(idx)
        self.prepareData()
        if self._view.model() != self:
            idx = self._view.model().mapFromSource(idx)
        self._view.setCurrentIndex(idx)

    def prepareData(self):
        self.beginResetModel()
        self._data = []
        self._data = Ledger.get_operations_sequence(self._begin, self._end, self._account)
        self.endResetModel()

    def delete_rows(self, rows):
        for row in rows:
            if (row >= 0) and (row < len(self._data)):
                try:
                    LedgerTransaction.get_operation(self._data[row]['otype'], self._data[row]['oid'], self._data[row]['opart']).delete()
                except IndexError as e:   # If row is already deleted (for example by another reference from 'opart')
                    if str(e) == LedgerTransaction.NoOpException:
                        pass
        self.prepareData()

    def assign_tag_to_rows(self, rows, tag_id):
        for row in rows:
            if (row >= 0) and (row < len(self._data)):
                LedgerTransaction.get_operation(self._data[row]['otype'], self._data[row]['oid'],self._data[row]['opart']).assign_tag(tag_id)
        self.prepareData()



from functools import partial
from datetime import datetime
import logging
import traceback

from PySide6.QtCore import Property, Slot
from PySide6.QtWidgets import QFileDialog, QMessageBox, QApplication

from jal.ui.ui_tax_export_widget import Ui_TaxWidget
from jal.ui.ui_flow_export_widget import Ui_MoneyFlowWidget
from jal.widgets.mdi import MdiWidget
from jal.widgets.helpers import ts2d, dt2ts
from jal.widgets.icons import JalIcon
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.db.peer import JalPeer
from jal.db.settings import JalSettings, FolderFor
from jal.data_export.taxes import TaxReport
from jal.data_export.taxes_flow import TaxesFlowRus
from jal.data_export.xlsx import XLSX
from jal.data_export.dlsg import DLSG


class TaxWidget(MdiWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_TaxWidget()
        self.ui.setupUi(self)

        self.ui.Country.clear()
        for x in TaxReport.countries:
            self.ui.Country.addItem(JalIcon.country_flag(TaxReport.countries[x]['flag']), TaxReport.countries[x]['name'])
        self.ui.Country.currentIndexChanged.connect(self.OnCountryChange)
        self.ui.Year.valueChanged.connect(self.OnYearChange)
        self.ui.Year.setValue(datetime.now().year - 1)   # Set previous year by default
        self.ui.XlsSelectBtn.pressed.connect(partial(self.OnFileBtn, 'XLS'))
        self.ui.DlsgSelectBtn.pressed.connect(partial(self.OnFileBtn, 'DLSG'))
        self.ui.SaveButton.pressed.connect(self.SaveReport)
        self.ui.Country.setCurrentIndex(TaxReport.RUSSIA)

    def OnCountryChange(self, item_id):
        if item_id == TaxReport.PORTUGAL:
            self.ui.RuBox.setVisible(False)
            self.ui.PtBox.setVisible(True)
        elif item_id == TaxReport.RUSSIA:
            self.ui.RuBox.setVisible(True)
            self.ui.PtBox.setVisible(False)
        else:
            raise ValueError("Selected item has no country handler in code")
        # Refresh and adjust MDI-window size
        if not self.parent() is None:
            self.parent().update()
            QApplication.processEvents()
            if not self.parent().isMaximized():  # Prevent size-change of maximized MDI
                self.parent().adjustSize()

    # Load account combobox with account names relevant for the given year
    def OnYearChange(self, year):
        self.ui.Account.clear()
        accounts = JalAccount.get_taxable_accounts(dt2ts(datetime(year, 1, 1)))
        for account in accounts:
            self.ui.Account.addItem(account.name(), account.id())

    # Displays tax widget in a given MDI area.
    # It is implemented as a separate static method in order to prevent unexpected object deletion
    @staticmethod
    def showInMDI(parent_mdi):
        parent_mdi.addSubWindow(TaxWidget(), maximized=False)

    @Slot()
    def OnFileBtn(self, type):
        if type == 'XLS':
            selector = (self.tr("Save tax reports to:"), self.tr("Excel files (*.xlsx)"), '.xlsx', self.ui.XlsFileName)
        elif type == 'DLSG':
            last_digit = self.year % 10
            selector = (self.tr("Save tax form to:"), self.tr(f"Tax form (*.dc{last_digit})"),
                        f".dc{last_digit}", self.ui.DlsgFileName)
        else:
            raise ValueError
        folder = JalSettings().getRecentFolder(FolderFor.Report, '.')
        filename = QFileDialog.getSaveFileName(self, selector[0], folder, selector[1])
        if filename[0]:
            if filename[1] == selector[1] and filename[0][-len(selector[2]):] != selector[2]:
                selector[3].setText(filename[0] + selector[2])
            else:
                selector[3].setText(filename[0])
            JalSettings().setRecentFolder(FolderFor.Report, filename[0])

    def getYear(self):
        return self.ui.Year.value()

    def getXlsFilename(self):
        return self.ui.XlsFileName.text()

    def getAccount(self):
        return self.ui.Account.currentData()

    def getDlsgState(self):
        return self.ui.DlsgGroup.isChecked()

    def getDslgFilename(self):
        return self.ui.DlsgFileName.text()

    def getBrokerAsIncomeName(self):
        return self.ui.DlsgIncomeSourceBroker.isChecked()

    def getDividendsOnly(self):
        return self.ui.DlsgDividendsOnly.isChecked()

    def getUseOneRate(self):
        return self.ui.Pt_OneCurrencyRate.isChecked()

    def getNoSettlement(self):
        return self.ui.NoSettlement.isChecked()

    year = Property(int, fget=getYear)
    xls_filename = Property(str, fget=getXlsFilename)
    account = Property(int, fget=getAccount)
    update_dlsg = Property(bool, fget=getDlsgState)
    dlsg_filename = Property(str, fget=getDslgFilename)
    dlsg_broker_as_income = Property(bool, fget=getBrokerAsIncomeName)
    dlsg_dividends_only = Property(bool, fget=getDividendsOnly)
    use_one_rate = Property(bool, fget=getUseOneRate)
    no_settelement = Property(bool, fget=getNoSettlement)

    @Slot()
    def SaveReport(self):
        if not self.account:
            QMessageBox().warning(self, self.tr("Data are incomplete"),
                                  self.tr("You haven't selected an account for tax report"), QMessageBox.Ok)
            return
        taxes = TaxReport.create_report(self.ui.Country.currentIndex())

        tax_report = taxes.prepare_tax_report(self.year, self.account,
                                              use_one_currency_rate=self.use_one_rate,
                                              use_settlement=(not self.no_settelement))
        if not tax_report:
            logging.warning(self.tr("Tax report is empty"))
            return

        reports_xls = XLSX(self.xls_filename)
        parameters = {
            "period": f"{ts2d(taxes.year_begin)} - {ts2d(taxes.year_end - 1)}",
            "account": f"{taxes.account.number()} ({JalAsset(taxes.account.currency()).symbol()})",
            "currency": JalAsset(taxes.account.currency()).symbol(),
            "broker_name": JalPeer(taxes.account.organization()).name(),
            "broker_iso_country": taxes.account.country().iso_code()
        }
        for section in tax_report:
            reports_xls.output_data(tax_report[section], taxes.report_template(section), parameters)
        reports_xls.save()
        logging.info(self.tr("Tax report was saved to file ") + f"'{self.xls_filename}'")

        if self.update_dlsg:
            tax_forms = DLSG(self.year, broker_as_income=self.dlsg_broker_as_income,
                             only_dividends=self.dlsg_dividends_only)
            tax_forms.update_taxes(tax_report, parameters)
            try:
                tax_forms.save(self.dlsg_filename)
                logging.info(self.tr("Tax report saved to file ") + f"'{self.dlsg_filename}'")
            except:
                logging.error(self.tr("Can't write tax form into file ") + f"'{self.dlsg_filename}'" +
                              f"\n{traceback.format_exc()}")


class MoneyFlowWidget(MdiWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_MoneyFlowWidget()
        self.ui.setupUi(self)

        self.ui.Year.setValue(datetime.now().year - 1)  # Set previous year by default
        self.ui.XlsSelectBtn.pressed.connect(self.OnFileBtn)
        self.ui.SaveButton.pressed.connect(self.SaveReport)

    # Displays tax widget in a given MDI area.
    # It is implemented as a separate static method in order to prevent unexpected object deletion
    @staticmethod
    def showInMDI(parent_mdi):
        parent_mdi.addSubWindow(MoneyFlowWidget(parent_mdi), maximized=False)

    @Slot()
    def OnFileBtn(self):
        selector = (self.tr("Save money flow report to:"), self.tr("Excel files (*.xlsx)"), '.xlsx', self.ui.XlsFileName)
        folder = JalSettings().getRecentFolder(FolderFor.Report, '.')
        filename = QFileDialog.getSaveFileName(self, selector[0], folder, selector[1])
        if filename[0]:
            if filename[1] == selector[1] and filename[0][-len(selector[2]):] != selector[2]:
                selector[3].setText(filename[0] + selector[2])
            else:
                selector[3].setText(filename[0])
            JalSettings().setRecentFolder(FolderFor.Report, filename[0])

    def getYear(self):
        return self.ui.Year.value()

    def getXlsFilename(self):
        return self.ui.XlsFileName.text()

    year = Property(int, fget=getYear)
    xls_filename = Property(str, fget=getXlsFilename)

    @Slot()
    def SaveReport(self):
        taxes_flow = TaxesFlowRus()
        flow_report = taxes_flow.prepare_flow_report(self.year)

        reports_xls = XLSX(self.xls_filename)
        parameters = {
            "period": f"{ts2d(taxes_flow.year_begin)} - {ts2d(taxes_flow.year_end - 1)}"
        }
        reports_xls.output_data(flow_report, "tax_rus_flow.json", parameters)
        reports_xls.save()

        logging.info(self.tr("Money flow report saved to file ") + f"'{self.xls_filename}'")
        self.close()



import logging
import platform
from enum import auto
import requests
from requests.exceptions import ConnectTimeout, ConnectionError
from PySide6.QtCore import QThread, QMutex
from jal import __version__


# Class that executes web-requests in a separate thread
# Request parameters are given in constructor and execution starts immediately after object creation
# Result of execution is available via data() method after thread completion
class WebRequest(QThread):
    GET = auto()        # Execute HTTP GET method
    POST = auto()       # Execute HTTP POST method with application/x-www-form-urlencoded
    POST_JSON = auto()  # Execute HTTP POST with JSON

    def __init__(self, operation, url, params=None, headers=None, binary=False):
        super().__init__()
        self._mutex = QMutex()
        self._data = ''
        self._operation = operation
        self._url = url
        self._params = params
        self._headers = headers
        self._binary = binary
        if not self.isRunning():
            self.start()

    def run(self):
        self._mutex.lock()
        url = self._url
        operation = self._operation
        params = self._params
        headers = self._headers
        binary = self._binary
        self._mutex.unlock()
        result = self._request(operation, url, params=params, headers=headers, binary=binary)
        self._mutex.lock()
        self._data = result
        self._mutex.unlock()

    def data(self):
        self._mutex.lock()
        data = self._data
        self._mutex.unlock()
        return data

    def _request(self, operation, url, params=None, headers=None, binary=False):
        session = requests.Session()
        session.headers['User-Agent'] = f"JAL/{__version__} ({platform.system()} {platform.release()})"
        if headers is not None:
            session.headers.update(headers)
        try:
            if operation == WebRequest.GET:
                response = session.get(url, params=params)
            elif operation == WebRequest.POST:
                response = session.post(url, data=params)
            elif operation == WebRequest.POST_JSON:
                response = session.post(url, json=params)
            else:
                assert False
        except ConnectTimeout:
            logging.error(self.tr("Timeout") + " URL {url}")
            return ''
        except ConnectionError as e:
            logging.error(self.tr("Error") + ", URL {url}\n{e}")
            return ''
        if response.status_code == 200:
            if binary:
                return response.content
            else:
                return response.text
        else:
            logging.error(self.tr("Failed") + f" [{response.status_code}] URL {url}\n{response.text}")
            return ''



from decimal import Decimal
from functools import partial
from PySide6.QtCore import Qt, Slot, QObject, QDateTime, QAbstractTableModel
from PySide6.QtGui import QFont
from jal.reports.reports import Reports
from jal.db.deposit import JalDeposit
from jal.db.helpers import localize_decimal
from jal.ui.reports.ui_term_deposits_report import Ui_TermDepositsReportWidget
from jal.widgets.delegates import FloatDelegate, TimestampDelegate
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "TermDepositsReport"


# ----------------------------------------------------------------------------------------------------------------------
class DepositsListModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [self.tr("Name"), self.tr("Start Date"), self.tr("End Date"), self.tr("Currency"),
                         self.tr("Initial amount"), self.tr("Accrued Interest"), self.tr("Planned End Value")]
        self._timestamp = 0
        self._view = parent_view
        self._data = []
        self._float_delegate = None
        self._timestamp_delegate = None
        self._initial_total = Decimal('0')
        self._accrued_total = Decimal('0')
        self._planned_total = Decimal('0')

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def headerWidth(self, section):
        return self._view.horizontalHeader().sectionSize(section)

    def data(self, index, role=Qt.DisplayRole, field=''):
        if role == Qt.DisplayRole:
            deposit = self._data[index.row()]
            return self.data_text(deposit, index.column())

    def data_text(self, deposit: JalDeposit, column: int):
        if column == 0:
            return deposit.name()
        if column == 1:
            return deposit.start_date()
        if column == 2:
            return deposit.end_date()
        if column == 3:
            return deposit.currency().symbol()
        if column == 4:
            return deposit.open_amount()
        if column == 5:
            return deposit.accrued_interest(self._timestamp)
        if column == 6:
            return deposit.close_amount()
        return ''

    def footerData(self, section: int, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if section ==0:
                return self.tr("Total")
            if section == 4:
                return localize_decimal(self._initial_total, precision=2)
            if section == 5:
                return localize_decimal(self._accrued_total, precision=2)
            if section == 6:
                return localize_decimal(self._planned_total, precision=2)
        elif role == Qt.FontRole:
            font = QFont()
            font.setBold(True)
            return font
        elif role == Qt.TextAlignmentRole:
            if section > 3:
                return Qt.AlignRight | Qt.AlignVCenter
            else:
                return Qt.AlignLeft | Qt.AlignVCenter
        return None

    def updateView(self, timestamp):
        if self._timestamp != timestamp:
            self._timestamp = timestamp
            self.prepareData()
            self.configureView()

    def prepareData(self):
        self.beginResetModel()
        self._data = JalDeposit.get_term_deposits(self._timestamp)
        self._initial_total = sum([x.open_amount() for x in self._data])
        self._accrued_total = sum([x.accrued_interest(self._timestamp) for x in self._data])
        self._planned_total = sum([x.close_amount() for x in self._data])
        self.endResetModel()

    def configureView(self):
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)
        self._float_delegate = FloatDelegate(2, allow_tail=False)
        self._timestamp_delegate = TimestampDelegate(display_format='%d/%m/%Y')
        self._view.setItemDelegateForColumn(1, self._timestamp_delegate)
        self._view.setItemDelegateForColumn(2, self._timestamp_delegate)
        self._view.setItemDelegateForColumn(4, self._float_delegate)
        self._view.setItemDelegateForColumn(5, self._float_delegate)
        self._view.setItemDelegateForColumn(6, self._float_delegate)
        self._view.setColumnWidth(0, 200)
        self._view.setColumnWidth(1, self._view.fontMetrics().horizontalAdvance("00/00/0000") * 1.1)
        self._view.setColumnWidth(2, self._view.fontMetrics().horizontalAdvance("00/00/0000") * 1.1)
        self._view.setColumnWidth(3, 100)
        self._view.setColumnWidth(4, 150)
        self._view.setColumnWidth(5, 150)
        self._view.setColumnWidth(6, 150)

# ----------------------------------------------------------------------------------------------------------------------
class TermDepositsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("Term deposits")
        self.window_class = "TermDepositsReportWindow"

# ----------------------------------------------------------------------------------------------------------------------
class TermDepositsReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_TermDepositsReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Term deposits")

        self.payments_model = DepositsListModel(self.ui.ReportTableView)
        self.ui.ReportTableView.setModel(self.payments_model)

        self.connect_signals_and_slots()

        current_time = QDateTime.currentDateTime()
        current_time.setTimeSpec(Qt.UTC)  # We use UTC everywhere so need to force TZ info
        self.ui.DepositsDate.setDateTime(current_time)

    def connect_signals_and_slots(self):
        self.ui.DepositsDate.dateChanged.connect(self.updateReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.ReportTableView.model()))

    @Slot()
    def updateReport(self):
        self.ui.ReportTableView.model().updateView(timestamp=self.ui.DepositsDate.date().endOfDay(Qt.UTC).toSecsSinceEpoch())



import logging
import re
from datetime import timezone
from jal.constants import PredefinedCategory
from jal.data_import.statement_xls import StatementXLS
from jal.data_import.statement import FOF, Statement_ImportError


JAL_STATEMENT_CLASS = "StatementVTB"

MAX_T_DELTA = 3  # Maximum allowed days between bond maturity and money transfer

# ----------------------------------------------------------------------------------------------------------------------
class StatementVTB(StatementXLS):
    AccountPattern = None  # to be set during validation
    PeriodPattern = None  # to be set during validation
    HeaderCol = 1
    money_section = "^Отчет об остатках денежных средств"
    money_columns = {
        "name": "Валюта",
        "cash_end": "Плановый",
    }
    asset_section = "^Отчет об остатках ценных бумаг"
    asset_columns = {
        "name": "Наименование ценной бумаги, \n№ гос. регистрации, ISIN",
        "currency": r"Валюта цены \n\(номинала для облигаций\)"
    }

    def __init__(self):
        super().__init__()
        self.name = self.tr("VTB Investments")
        self.icon_name = "vtb.ico"
        self.filename_filter = self.tr("VTB statement (*.xlsx)")
        self.account_end_balance = {}
        self.asset_withdrawal = []

    def _validate(self):
        shift = 0   # Check how header is shifted to the right
        for i in range(self._statement.shape[1]):
            if self._statement[i][4].startswith("Клиент"):  # Starting cell is found
                shift = 1
                continue
            if self._statement[i][4].startswith("ИНН"):   # Next header reached - something went wrong
                shift = 0
                break
            if not shift:
                continue
            if self._statement[i][4]:  # Client name found - we may break here
                break
            shift += 1
        if not shift:
            raise Statement_ImportError(self.tr("Can't determine VTB statement header format"))
        shift = shift - 6
        self.AccountPattern = (7 + shift, 7, None)
        self.PeriodPattern = (5 + shift, 1, r"Отчет Банка ВТБ \(ПАО\) за период с (?P<S>\d\d\.\d\d\.\d\d\d\d) по (?P<E>\d\d\.\d\d\.\d\d\d\d) о сделках, .*")
        super()._validate()

    def _strip_unused_data(self):
        for asset in self._data[FOF.ASSETS]:
            self.drop_extra_fields(asset, ['broker_name'])

    def _load_currencies(self):
        cnt = 0
        row, headers = self.find_section_start(self.money_section, self.money_columns, header_height=3)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row].startswith('Сумма') or self._statement[self.HeaderCol][row] == '':
                break
            currency = self._statement[headers['name']][row]
            try:
                code = self.currency_substitutions[currency]
            except KeyError:
                code = currency
            self.currency_id(code)
            self.account_end_balance[code] = self._statement[headers['cash_end']][row]
            cnt += 1
            row += 1
        logging.info(self.tr("Account currencies loaded: ") + f"{cnt}")

    # Planned money amount already loaded in _load_currencies(). Here it is only put in account data
    def _load_money(self):
        for currency in self.account_end_balance:
            account_id = self._find_account_id(self._account_number, currency)
            account = [x for x in self._data[FOF.ACCOUNTS] if x['id'] == account_id][0]
            account["cash_end"] = self.account_end_balance[currency]

    def _load_assets(self):
        AssetPattern = r"^(?P<name>.*), (?P<reg_number>.*), (?P<isin>.*)$"
        cnt = 0
        row, headers = self.find_section_start(self.asset_section, self.asset_columns)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row].startswith('ИТОГО') or self._statement[self.HeaderCol][row] == '':
                break
            if self._statement[self.HeaderCol][row].startswith(('АКЦИЯ', 'ОБЛИГАЦИЯ', 'ПАЙ')):
                row += 1  # This is an asset class separator - move to the next row
                continue
            asset_name = self._statement[headers['name']][row]
            parts = re.match(AssetPattern, asset_name, re.IGNORECASE)
            if parts is None:
                raise Statement_ImportError(self.tr("Can't parse asset name ") + f"'{asset_name}'")
            asset_data = parts.groupdict()
            if len(asset_data) != AssetPattern.count("(?P<"):  # check that expected number of groups was matched
                raise Statement_ImportError(self.tr("Asset name miss some data ") + f"'{asset_name}'")
            currency = self._statement[headers['currency']][row]
            try:
                currency_code = self.currency_substitutions[currency]
            except KeyError:
                currency_code = currency
            asset_id = self.asset_id({'isin': asset_data['isin'], 'reg_number': asset_data['reg_number'],
                                      'currency': self.currency_id(currency_code), 'search_offline': True,
                                      'search_online': "MOEX"})
            asset = self._find_in_list(self._data[FOF.ASSETS], 'id', asset_id)
            asset['broker_name'] = asset_name
            cnt += 1
            row += 1
        logging.info(self.tr("Securities loaded: ") + f"{cnt}")

    def _strip_unused_data(self):
        for asset in self._data[FOF.ASSETS]:
            self.drop_extra_fields(asset, ['broker_name'])

    def _load_deals(self):
        self._load_deals_main_market()
        self._load_deals_derivatives_market()
        self._load_deals_currency_exchange()

    def _load_deals_main_market(self):
        cnt = 0
        columns = {
            "asset_name": r"Наименование ценной бумаги, \n№ гос\. Регистрации, ISIN",
            "number": "№ сделки",
            "datetime": "Дата и время заключения сделки",
            "B/S": "Вид сделки",
            "price": r"Цена\n\(% для облигаций\)",
            "currency": r"Валюта цены\n \(номинала для облигаций\)",
            "qty": r"Количество \n\(шт\)",
            "amount": r"Сумма сделки в валюте расчетов\n \(с учетом НКД для облигаций\)  ",
            "accrued_int": "НКД\nпо сделке в валюте расчетов",
            "settlement": " Плановая дата поставки",
            "fee1": "Комиссия Банка за расчет по сделке",
            "fee2": "Комиссия Банка за заключение сделки"
        }
        row, headers = self.find_section_start(r"^Заключенные в отчетном периоде сделки с ценными бумагами", columns)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            asset_name = self._statement[headers['asset_name']][row]
            assets = [x for x in self._data[FOF.ASSETS] if x.get('broker_name') == asset_name]
            if len(assets) != 1:
                raise Statement_ImportError(self.tr("No asset match in deals for ") + f"'{asset_name}'")
            asset_id = assets[0]['id']
            if self._statement[headers['B/S']][row] == 'Покупка':
                qty = self._statement[headers['qty']][row]
                bond_interest = -self._statement[headers['accrued_int']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                qty = -self._statement[headers['qty']][row]
                bond_interest = self._statement[headers['accrued_int']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue
            deal_number = self._statement[headers['number']][row]
            price = self._statement[headers['price']][row]
            currency = self._statement[headers['currency']][row]
            fee = self._statement[headers['fee1']][row] + self._statement[headers['fee2']][row]
            amount = self._statement[headers['amount']][row]
            if abs(abs(price * qty) - amount) >= self.RU_PRICE_TOLERANCE:
                price = abs((amount - bond_interest) / qty)
            timestamp = int(self._statement[headers['datetime']][row].replace(tzinfo=timezone.utc).timestamp())
            settlement = int(self._statement[headers['settlement']][row].replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, currency)
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": deal_number, "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            if bond_interest != 0:
                new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
                payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                           "number": deal_number, "asset": asset_id, "amount": bond_interest, "description": "НКД"}
                self._data[FOF.ASSET_PAYMENTS].append(payment)
            cnt += 1
            row += 1
        logging.info(self.tr("Stock trades loaded: ") + f"{cnt}")

    def _load_deals_derivatives_market(self):
        cnt = 0
        columns = {
            "symbol": " Фьючерсный контракт / опцион, код",
            "number": "№ сделки",
            "datetime": "Дата и время заключения сделки",
            "B/S": "Вид сделки",
            "price": r"Цена контракта /\n размер премии \n\(пункты\)",
            "qty": r"Количество \n\(шт\)",
            "fee1": "Комиссия Банка за расчет по сделке",
            "fee2": "Комиссия Банка за заключение сделки"
        }
        row, headers = self.find_section_start(r"^Сделки с Производными финансовыми инструментами в отчетном периоде", columns)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            asset_id = self.asset_id({'symbol': self._statement[headers['symbol']][row], 'search_online': "MOEX"})
            if self._statement[headers['B/S']][row] == 'Покупка':
                qty = self._statement[headers['qty']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                qty = -self._statement[headers['qty']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue
            deal_number = self._statement[headers['number']][row]
            price = self._statement[headers['price']][row]
            fee = self._statement[headers['fee1']][row] + self._statement[headers['fee2']][row]
            timestamp = int(self._statement[headers['datetime']][row].replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, 'RUR')
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": deal_number, "timestamp": timestamp,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            cnt += 1
            row += 1
        logging.info(self.tr("Derivative trades loaded: ") + f"{cnt}")

    def _load_deals_currency_exchange(self):
        SymbolPattern = r"^(?P<A>[A-Z]{3})(?P<B>[A-Z]{3})(_TO[D|M])?$"
        cnt = 0
        columns = {
            "symbol": "Финансовый инструмент",
            "datetime": "Дата и время заключения сделки",
            "B/S": "Вид сделки",
            "number": "№ сделки",
            "deposit": r"Количество \(шт\)",
            "withdrawal": "Сумма сделки в валюте расчетов",
            "fee1": "Комиссия Банка за расчет по сделке",
            "fee2": "Комиссия Банка за заключение сделки",
            "fee_currency": "Валюта расчетов",
            "description": "Комментарий"
        }
        row, headers = self.find_section_start(r"^Заключенные в отчетном периоде сделки с иностранной валютой", columns)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            parts = re.match(SymbolPattern, self._statement[headers['symbol']][row], re.IGNORECASE)
            if parts is None:
                raise Statement_ImportError(self.tr("Can't parse currency exchange symbol") + f"'{self._statement[headers['symbol']][row]}'")
            symbols_data = parts.groupdict()
            if self._statement[headers['B/S']][row] == 'Покупка':
                symbols_data['A'], symbols_data['B'] = symbols_data['B'], symbols_data['A']
            elif self._statement[headers['B/S']][row] == 'Продажа':
                pass
            else:
                row += 1
                logging.warning(self.tr("Unknown currency trade type: ") + self._statement[headers['B/S']][row])
                continue
            timestamp = int(self._statement[headers['datetime']][row].replace(tzinfo=timezone.utc).timestamp())
            number = self._statement[headers['number']][row]
            account_to = self._find_account_id(self._account_number, symbols_data['B'])
            account_from = self._find_account_id(self._account_number, symbols_data['A'])
            account_fee = self._find_account_id(self._account_number, self._statement[headers['fee_currency']][row])
            withdrawal = self._statement[headers['withdrawal']][row]
            deposit = self._statement[headers['deposit']][row]
            fee = self._statement[headers['fee1']][row] + self._statement[headers['fee2']][row]
            description = self._statement[headers['description']][row]
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
            transfer = {"id": new_id, "account": [account_from, account_to, account_fee], "timestamp": timestamp, "number" : number,
                        "asset": [self.currency_id(symbols_data['A']), self.currency_id(symbols_data['B'])],
                        "withdrawal": withdrawal, "deposit": deposit, "fee": fee, "description": description}
            self._data[FOF.TRANSFERS].append(transfer)
            cnt += 1
            row += 1
        logging.info(self.tr("Currency trades loaded: ") + f"{cnt}")

    def _load_asset_transactions(self):
        cnt = 0
        columns = {
            "asset_name": "Наименование ценной бумаги, № гос. Регистрации, ISIN",
            "date": "Дата операции",
            "qty": r"Количество \n\(шт\.\)",
            "type": "Тип операции"
        }
        operations = {
            'Покупка': None,
            'Продажа': None,
            'Погашение ЦБ': self.asset_cancellation,
        }
        row, headers = self.find_section_start("^Движение ценных бумаг", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':  # Stop if there are no next date available
                break
            operation = self._statement[headers['type']][row]
            if operation not in operations:
                raise Statement_ImportError(self.tr("Unsuppported asset transaction ") + f"'{operation}'")
            asset_name = self._statement[headers['asset_name']][row]   # FIXME the same piece of code is in load_deals
            assets = [x for x in self._data[FOF.ASSETS] if x.get('broker_name') == asset_name]
            if len(assets) != 1:
                raise Statement_ImportError(self.tr("No asset match in asset transactions ") + f"'{asset_name}'")
            asset_id = assets[0]['id']
            timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            try:
                qty = float(self._statement[headers['qty']][row])
            except ValueError:
                raise Statement_ImportError(self.tr("Failed to convert asset amount ") + f"'{self._statement[headers['qty']][row]}'")
            if operations[operation] is not None:
                operations[operation](timestamp, asset_id, qty)
            cnt += 1
            row += 1
        logging.info(self.tr("Asset transactions loaded: ") + f"{cnt}")

    def asset_cancellation(self, timestamp, asset_id, qty):
        # Statement has negative value for cancellation - will be used to create sell trade
        self.asset_withdrawal.append({"timestamp": timestamp, "asset": asset_id, "quantity": qty})

    def _load_cash_transactions(self):
        cnt = 0
        columns = {
            "date": "Дата",
            "type": "Тип операции",
            "amount": "Сумма",
            "currency": "Валюта",
            "description": "Комментарий"
        }
        operations = {
            'Зачисление денежных средств': self.transfer_in,
            'Списание денежных средств': self.transfer_out,
            'Купонный доход': self.interest,
            'Погашение ценных бумаг': self.bond_maturity,
            'Сальдо расчётов по сделкам с ценными бумагами': None,  # These operations are results of trades
            'Вознаграждение Брокера': self.broker_fee,
            'Дивиденды': self.dividend,
            'Сальдо расчётов по сделкам с иностранной валютой': None,  # These operation are results of currency exchange
            'Перевод денежных средств': None,   # TODO - to be implemented
            'Вариационная маржа': None
        }
        row, headers = self.find_section_start("^Движение денежных средств", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':  # Stop if there are no next date available
                break
            operation = self._statement[headers['type']][row]
            if operation == '':  # Skip market header as of now
                row += 1
                continue
            if operation not in operations:
                raise Statement_ImportError(self.tr("Unsuppported cash transaction ") + f"'{operation}'")
            timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            amount = self._statement[headers['amount']][row]
            description = self._statement[headers['description']][row]
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            if operations[operation] is not None:
                operations[operation](timestamp, account_id, amount, description)
            cnt += 1
            row += 1
        logging.info(self.tr("Cash operations loaded: ") + f"{cnt}")

    def transfer_in(self, timestamp, account_id, amount, description):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_out(self, timestamp, account_id, amount, description):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_id, 0, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": -amount, "deposit": -amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def interest(self, timestamp, account_id, amount, description):
        BondInterestPattern = r"^Куп\. дох\. по обл\. .* (?P<reg_number>\S*), \S*\..*$"
        parts = re.match(BondInterestPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse bond interest description ") + f"'{description}'")
        interest_data = parts.groupdict()
        asset_id = self._find_in_list(self._data[FOF.ASSETS_DATA], 'reg_number', interest_data['reg_number'])['asset']
        new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                   "asset": asset_id, "amount": amount, "description": description}
        self._data[FOF.ASSET_PAYMENTS].append(payment)

    def dividend(self, timestamp, account_id, amount, description):
        DividendPattern = r"^Дивиденды .* (?P<reg_number>\S*), .*. Удержан налог в размере (?P<tax>\d+\.\d\d) руб.$"
        parts = re.match(DividendPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse dividend description ") + f"'{description}'")
        dividend_data = parts.groupdict()
        asset_id = self._find_in_list(self._data[FOF.ASSETS_DATA], 'reg_number', dividend_data['reg_number'])['asset']
        try:
            tax = float(dividend_data['tax'])
            amount += tax
        except ValueError:
            raise Statement_ImportError(self.tr("Failed to convert dividend tax ") + f"'{description}'")
        new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        payment = {"id": new_id, "type": FOF.PAYMENT_DIVIDEND, "account": account_id, "timestamp": timestamp,
                   "asset": asset_id, "amount": amount, "tax": tax, "description": description}
        self._data[FOF.ASSET_PAYMENTS].append(payment)

    def bond_maturity(self, timestamp, account_id, amount, description):
        MaturityPattern = r"^Ден\.ср-ва от погаш\. номин\.ст-ти обл\. .* (?P<reg_number>\S*), .* Налог не удерживается\.$"
        parts = re.match(MaturityPattern, description, re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse bond maturity description ") + f"'{description}'")
        bond_maturity = parts.groupdict()
        asset_id = self._find_in_list(self._data[FOF.ASSETS_DATA], 'reg_number', bond_maturity['reg_number'])['asset']
        match = [x for x in self.asset_withdrawal if x['asset'] == asset_id and (timestamp - x['timestamp']) <= MAX_T_DELTA*86400]
        if not match:
            breakpoint()
            raise Statement_ImportError(self.tr("Can't find asset cancellation record for ") + f"'{description}'")
        if len(match) != 1:
            raise Statement_ImportError(self.tr("Multiple asset cancellation match for ") + f"'{description}'")
        asset_cancel = match[0]
        qty = asset_cancel['quantity']
        price = abs(amount / qty)  # Price is always positive
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
        trade = {"id": new_id, "timestamp": timestamp, "settlement": timestamp, "account": account_id,
                 "asset": asset_id, "quantity": qty, "price": price, "fee": 0.0, "note": description}
        self._data[FOF.TRADES].append(trade)

    def broker_fee(self, timestamp, account_id, amount, description):
        if description.startswith("Комиссия банка за заключение сделок") \
            or description.startswith("Комиссия за брокерские услуги по проведению расчетов по заключенным сделкам"):
            return  # Skip these operations as they are already included in trade fees
        new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        fee = {"id": new_id, "peer": 0, "account": account_id, "timestamp": timestamp,
               'lines': [{'amount': amount, 'category': -PredefinedCategory.Fees, 'description': description}]}
        self._data[FOF.INCOME_SPENDING].append(fee)



# File to fetch asset data from MOEX (except quotes)

import logging
import re
import json
from datetime import datetime, timezone
from PySide6.QtWidgets import QApplication
from jal.widgets.helpers import is_english
from jal.constants import PredefinedAsset
from jal.net.web_request import WebRequest

# ===================================================================================================================
class MOEX:
    mapping = {
        'SECID': 'symbol',
        'NAME': 'name',
        'SHORTNAME': 'short_name',
        'ISIN': 'isin',
        'REGNUMBER': 'reg_number',
        'FACEVALUE': 'principal',
        'MATDATE': 'expiry',
        'LSTDELDATE': 'expiry',
        'GROUP': 'type'
    }
    asset_type = {
        'stock_shares': PredefinedAsset.Stock,
        'stock_dr': PredefinedAsset.Stock,
        'stock_bonds': PredefinedAsset.Bond,
        'stock_etf': PredefinedAsset.ETF,
        'stock_ppif': PredefinedAsset.ETF,
        'futures_forts': PredefinedAsset.Derivative,
        'currency_metal': PredefinedAsset.Commodity
    }
    FuturesPattern = r'^(?P<base_asset>\w+)-\d{1,2}\.\d{2}$'

    def __init__(self):
        pass

    # Searches for asset info on http://www.moex.com/ identified by info provided:
    # kwargs - dict with one of the following keys: 'reg_number', 'isin' or 'name'
    # Returns symbol ('secid' of MOEX) if asset was found and empty string otherwise
    def find_asset(self, **kwargs) -> str:
        asset_data = []
        if reg_nbr := kwargs.get('reg_number', ''):
            asset_data = self.__lookup_asset('reg_number', reg_nbr)
        if not asset_data and (isin := kwargs.get('isin', '')):
            asset_data = self.__lookup_asset('isin', isin)
        if not asset_data and (name := kwargs.get('name', '')):
            asset_data = self.__lookup_asset('name', name)
            if not asset_data and (match := re.match(self.FuturesPattern, name)):
                asset_data = self.__lookup_futures(name, match.group('base_asset')) # not found by name - try extended search if it is a futures contract
        if asset_data and 'secid' in asset_data:
            return asset_data['secid']
        else:
            return ''

    # Searches for asset via MOEX main search: https://iss.moex.com/iss/securities.json
    # key - key to search for (isin, reg_number, name)
    # search_value - value to search for, like ISIN, registration number or asset name
    # Returns dictionary with available asset data or empty dictionary if asset not found
    def __lookup_asset(self, key, search_value) -> dict:
        # Define which columns to search for given value
        fields = {'isin': ['isin'], 'reg_number': ['regnumber'], 'name': ['name', 'shortname']}
        if key not in fields:
            return {}
        request = WebRequest(WebRequest.GET, "https://iss.moex.com/iss/securities.json",
                             params={'q': search_value, 'iss.meta': 'off', 'limit': '10'})
        while request.isRunning():
            QApplication.processEvents()
        moex_search_data = json.loads(request.data())
        securities = moex_search_data['securities']
        data = []
        for field in fields[key]:
            key_col = securities['columns'].index(field)
            if data := [x for x in securities['data'] if x[key_col] == search_value or x[key_col] is None]:
                break
        data = self.__validate_search_data(data, securities['columns'], key, search_value)
        if not data:
            return {}
        return dict(zip(securities['columns'], data[0]))

    def __lookup_futures(self, name, base_asset) -> dict:
        request = WebRequest(WebRequest.GET, "https://iss.moex.com/iss/statistics/engines/futures/markets/forts/series.json",
                             params={'asset_code': base_asset, 'show_expired': '1'})
        while request.isRunning():
            QApplication.processEvents()
        moex_search_data = json.loads(request.data())
        futures = moex_search_data['series']
        key_col = futures['columns'].index('name')
        data = [x for x in futures['data'] if x[key_col] == name or x[key_col] is None]
        data = self.__validate_search_data(data, futures['columns'], 'name', name)
        if not data:
            return {}
        return dict(zip(futures['columns'], data[0]))

    def __validate_search_data(self, data, columns, key, search_value) -> list:
        if not data:
            logging.info(QApplication.translate("MOEX", "No MOEX assets found for ") + f"{key} = {search_value}")
            return []
        if len(data) > 1:  # remove not traded assets if there are any outdated doubles present
            data = [x for x in data if x[columns.index('is_traded')] == 1]
            if not data:
                logging.info(QApplication.translate("MOEX", "Multiple MOEX non-traded assets found for ") + f"{key} = {search_value}: {data}")
                return []
        if len(data) > 1:  # still have multiple matches for the same asset data
            logging.info(QApplication.translate("MOEX", "Multiple MOEX assets found for ") + f"{key} = {search_value}: {data}")
            return []
        return data

    # Fetches asset information from http://moex.com/ for asset identified by kwargs
    # kwargs - dict that can contains:
    # - one of the following mandatory keys: 'symbol', 'isin', 'reg_number'
    # - optional 'currency' key to get info in specific currency
    # - optional 'special' key to get moex specific 'engine', 'market' and 'board' parameters of the asset
    # Returns dictionary with asset data or empty dictionary if asset not found
    def asset_info(self, **kwargs) -> dict:
        data = {}
        if 'symbol' in kwargs and not is_english(kwargs['symbol']):
            del kwargs['symbol']
        currency = kwargs['currency'] if 'currency' in kwargs else ''
        # First try to load with symbol or isin from asset details API
        if 'symbol' in kwargs:
            data = self.__get_info(kwargs['symbol'], currency=currency)
        if not data and 'isin' in kwargs:
            data = self.__get_info(kwargs['isin'], currency=currency)
        # If not found try to use search API with regnumber or isin
        if not data and 'reg_number' in kwargs:
            data = self.__get_info(MOEX().find_asset(reg_number=kwargs['reg_number']))
        if not data and 'isin' in kwargs:
            data = self.__get_info(MOEX().find_asset(isin=kwargs['isin']))
        if 'special' not in kwargs:
            for key in ['engine', 'market', 'board']:
                try:
                    del data[key]
                except KeyError:
                    pass
        return data

    def __get_info(self, asset_code, currency='') -> dict:
        asset = {}
        if not asset_code:
            return asset
        request = WebRequest(WebRequest.GET, f"http://iss.moex.com/iss/securities/{asset_code}.json")
        while request.isRunning():
            QApplication.processEvents()
        moex_data = json.loads(request.data())
        boards = [dict(zip(moex_data['boards']['columns'], x)) for x in moex_data['boards']['data']]
        if not boards:
            return asset
        for row in moex_data['description']['data']:
            value = dict(zip(moex_data['description']['columns'], row))
            asset[value['name']] = value['value']
        asset = {self.mapping[x]: asset[x] for x in asset if x in self.mapping}  # map keys to internal names
        if 'isin' in asset:  # replace symbol with short name if we have isin in place of symbol
            asset['symbol'] = asset['short_name'] if asset['symbol'] == asset['isin'] else asset['symbol']
        if 'short_name' in asset:
            del asset['short_name']  # drop short name as we won't use it further
        if 'principal' in asset:  # Convert principal into float if possible or drop otherwise
            try:
                asset['principal'] = float(asset['principal'])
            except ValueError:
                del asset['principal']
        if 'expiry' in asset:  # convert YYYY-MM-DD into timestamp
            date_value = datetime.strptime(asset['expiry'], "%Y-%m-%d")
            asset['expiry'] = int(date_value.replace(tzinfo=timezone.utc).timestamp())
        try:
            asset['type'] = self.asset_type[asset['type']]
        except KeyError:
            logging.error(QApplication.translate("MOEX", "Unsupported MOEX security type: ") + f"{asset['type']}")
            return {}
        primary_board = [x for x in boards if "is_primary" in x and x["is_primary"] == 1]
        if primary_board:
            assert len(primary_board) == 1, "Unexpected multiple primary boards"
            board_id = primary_board[0]['boardid']
            if primary_board[0]['currencyid'] != currency and primary_board[0]['market'] != 'bonds':
                currency_suffix = {'USD': 'D', 'EUR': 'E', 'CNY': 'Y'}
                board_id = board_id[:-1] + currency_suffix[currency] if currency in currency_suffix else board_id
            board = [x for x in boards if 'boardid' in x and x['boardid'] == board_id]
            if board:
                asset.update({'engine': board[0]['engine'],
                              'market': board[0]['market'],
                              'board': board[0]['boardid']})
        return asset
